"""
    Protein.WypMassRatios.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from Bio.PDB import PDBParser, MMCIFParser, CaPPBuilder
from Bio.SeqUtils.ProtParam import ProteinAnalysis
from Bio.Data.IUPACData import protein_weights

pdb_parser = PDBParser()
cif_parser = MMCIFParser()
polypeptide_builder = CaPPBuilder()

def get_amino_acids_props(pdb_file):
    parser = cif_parser if pdb_file[-4:] == '.cif' else pdb_parser
    structure = parser.get_structure("XXXX", pdb_file)

    result_dict = {}
    for polypeptide in polypeptide_builder.build_peptides(structure):
        seq = polypeptide.get_sequence()
        analyzed_seq = ProteinAnalysis(str(seq))
        count_dict = analyzed_seq.count_amino_acids()
        for k, n in count_dict.items():
            v = result_dict.get(k, 0)
            v += n*protein_weights[k]
            result_dict[k] = v

    # print(result_dict)
    total = np.sum(list(result_dict.values()))
    props_dict = {}
    for k, v in result_dict.items():
        props_dict[k] = v/total

    return props_dict

def compute_mass_ratios(pdb_file):
    props_dict = get_amino_acids_props(pdb_file)
    w = props_dict['W']
    y = props_dict['Y']
    p = props_dict['P']
    return y/w, p/w

def create_mass_ratios_file(pdb_folder):
    from glob import glob

    ratio_file = os.path.join(pdb_folder, "wyp_ratios.csv")
    fh = open(ratio_file, "w")

    for path in glob(pdb_folder + r"\*.pdb"):
        print(path)
        ratios = compute_mass_ratios(path)
        folder, file = os.path.split(path)
        fh.write(",".join([file, *["%.3g" % r for r in ratios]]) + "\n")

    fh.close()

def estimate_coefficients():
    """
             
    1ova.pdb,2.96,2.63  0.981,0.552
    1ald.pdb,3.84,3.57  0.960,0.433
    2glk.pdb,1.33,1.79  0.965,0.566
    3v03.pdb,8.87,7.89  1.01, 0.611
    """
    X = [
          [0.981, 0.552],
          [0.960, 0.433],
          [0.965, 0.566],
          [1.01,  0.611],
          ]

    Y = [
          [2.96, 2.63],
          [3.84, 3.57],
          [1.33, 1.79],
          [8.87, 7.89],
          ]

if __name__ == '__main__':
    pdb_folder = r"D:\TODO\20231218\wyf-ratios\PDB"
    create_mass_ratios_file(pdb_folder)



# coding: utf-8
"""
    ZhiyaZuoGmm.py.

    adapted from:
        Expectation Maximization
        https://zhiyzuo.github.io/EM/

        or

        https://github.com/sohaib730/Gaussian-Mixture-Model

"""
import numpy as np
import numpy.random as rd
import scipy as sp
import seaborn as sns
from matplotlib import pyplot as plt
try:
    # for numba 1.49 or later
    from numba.core.decorators import jit
except:
    from numba.decorators import jit

class GMM(object):
    def __init__(self, X, k=2, max_iter=1000):
        # dimension
        X = np.asarray(X)
        print('X.shape=', X.shape)
        self.m, self.n = X.shape
        self.data = X.copy()
        # number of mixtures
        self.k = k
        self.max_iter = max_iter

    def _init(self):
        # init mixture means/sigmas
        if False:
            self.mean_arr = np.asmatrix(np.random.random((self.k, self.n)))
            self.sigma_arr = np.array([np.asmatrix(np.identity(self.n)) for i in range(self.k)])
        else:
            # improved procedure from MatsukenGmm.py.
            mu_list = []
            lim_list = []
            for d in range(self.n):
                max_x, min_x = np.max(self.data[:,d]), np.min(self.data[:,d])
                mu_list.append(rd.uniform(low=min_x, high=max_x, size=self.k))
            self.mean_arr = np.asmatrix(np.array(mu_list).T)
            self.sigma_arr = np.array([np.asmatrix(np.identity(self.n)*50) for i in range(self.k)])
        self.phi = np.ones(self.k)/self.k
        self.w = np.asmatrix(np.empty((self.m, self.k), dtype=float))
        print("init mean_arr=", self.mean_arr)
        print("init sigma_arr=", self.sigma_arr)
    
    def fit(self, tol=1e-4):
        self._init()
        num_iters = 0
        ll = 1
        previous_ll = 0
        print('self.max_iter=', self.max_iter)
        while(ll-previous_ll > tol):
            previous_ll = self.loglikelihood()
            self._fit()
            num_iters += 1
            if num_iters > self.max_iter:
                break
            ll = self.loglikelihood()
            print('Iteration %d: log-likelihood is %.6f'%(num_iters, ll))
        print('Terminate at %d-th iteration:log-likelihood is %.6f'%(num_iters, ll))
    
    def loglikelihood(self):
        ll = 0
        for i in range(self.m):
            tmp = 0
            for j in range(self.k):
                #print(self.sigma_arr[j])
                tmp += sp.stats.multivariate_normal.pdf(self.data[i, :], 
                                                        self.mean_arr[j, :].A1, 
                                                        self.sigma_arr[j, :]) *\
                       self.phi[j]
            ll += np.log(tmp) 
        return ll
    
    def _fit(self):
        self.e_step()
        self.m_step()

    # @jit(nopython=True)
    def e_step(self):
        # calculate w_j^{(i)}
        for i in range(self.m):
            den = 0
            for j in range(self.k):
                num = sp.stats.multivariate_normal.pdf(self.data[i, :], 
                                                       self.mean_arr[j].A1, 
                                                       self.sigma_arr[j]) *\
                      self.phi[j]
                den += num
                self.w[i, j] = num
            self.w[i, :] /= den
            wi_err = self.w[i, :].sum() - 1
            # print('wi_err=', wi_err)
            assert wi_err < 1e-4

    def m_step(self):
        for j in range(self.k):
            const = self.w[:, j].sum()
            self.phi[j] = 1/self.m * const
            _mu_j = np.zeros(self.n)
            _sigma_j = np.zeros((self.n, self.n))
            for i in range(self.m):
                _mu_j += (self.data[i, :] * self.w[i, j])
                _sigma_j += self.w[i, j] * ((self.data[i, :] - self.mean_arr[j, :]).T * (self.data[i, :] - self.mean_arr[j, :]))
                #print((self.data[i, :] - self.mean_arr[j, :]).T * (self.data[i, :] - self.mean_arr[j, :]))
            self.mean_arr[j] = _mu_j / const
            self.sigma_arr[j] = _sigma_j / const
        #print(self.sigma_arr)

def blog_demo():
    X = np.random.multivariate_normal([0, 3], [[0.5, 0], [0, 0.8]], 20)
    X = np.vstack((X, np.random.multivariate_normal([20, 10], np.identity(2), 50)))
    X.shape
    gmm = GMM(X)
    gmm.fit()

    print('mean', gmm.mean_arr)
    print('sigma', gmm.sigma_arr)
    print('phi', gmm.phi)

class GmmAdaptor:
    def __init__(self, k, max_iter=1000):
        self.k = k
        self.max_iter= max_iter

    def fit(self, X):

        while True:
            try:
                gmm = GMM(X, k=self.k, max_iter=self.max_iter)
                gmm.fit()
                break
            except AssertionError:
                # bad init. try until it succeeds.
                # this error can be avoided by improving init with kmeans.
                print('retry on AssertionError')
                continue

        print('mean', gmm.mean_arr)
        print('sigma', gmm.sigma_arr)
        print('phi', gmm.phi)

        self.weights_ = gmm.phi
        self.means_ = np.asarray(gmm.mean_arr).flatten()
        self.covariances_ = gmm.sigma_arr.flatten()



"""
    Solvers.UltraNest.CustomLivePointsWidget.py

    adapted from ultranest.viz.LivePointsWidget

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from ultranest.viz import round_parameterlimits, clusteridstrings, spearman

class CustomLivePointsWidget:
    def __init__(self, default_callback=None):
        self.grid = None
        self.label = None
        self.laststatus = None
        self.default_callback = default_callback
    
    def initialize(self, paramnames, width):
        from importlib import reload
        import Solvers.UltraNest.GridspecLayoutMpl
        reload(Solvers.UltraNest.GridspecLayoutMpl)
        from Solvers.UltraNest.GridspecLayoutMpl import GridspecLayoutMpl
        self.grid = GridspecLayoutMpl(len(paramnames), width + 3)
        self.laststatus = []
        for a, paramname in enumerate(paramnames):
            self.laststatus.append('*' * width)

    def __call__(self, points, info, region, transformLayer, region_fresh=False):
        if self.default_callback is not None:
            self.default_callback(points, info, region, transformLayer, region_fresh=region_fresh)

        p = points['p']
        paramnames = info['paramnames']

        plo = p.min(axis=0)
        phi = p.max(axis=0)
        plo_rounded, phi_rounded, paramformats = round_parameterlimits(plo, phi, paramlimitguess=info.get('paramlims'))

        width = 50

        if self.grid is None:
            self.initialize(paramnames, width)
        
        with np.errstate(invalid="ignore"):
            indices = ((p - plo_rounded) * width / (phi_rounded - plo_rounded).reshape((1, -1))).astype(int)
        indices[indices >= width] = width - 1
        indices[indices < 0] = 0
        ndim = len(plo)

        clusterids = transformLayer.clusterids % len(clusteridstrings)
        nmodes = transformLayer.nclusters
        labeltext = ("Mono-modal" if nmodes == 1 else "Have %d modes" % nmodes) + \
            (" | Volume: ~exp(%.2f) " % region.estimate_volume()) + ('*' if region_fresh else ' ') + \
            " | Expected Volume: exp(%.2f)" % info['logvol'] + \
            ('' if 'order_test_correlation' not in info else
             (" | Quality: correlation length: %d (%s)" % (info['order_test_correlation'], '+' if info['order_test_direction'] >= 0 else '-'))
             if np.isfinite(info['order_test_correlation']) else " | Quality: ok")

        if info.get('stepsampler_info', {}).get('num_logs', 0) > 0:
            stepsampler_info = dict(info['stepsampler_info'])
            stepsampler_info['frac_far_enough'] *= 100
            if 'mean_distance' in stepsampler_info:
                labeltext += (
                    "<br/>"
                    'Step sampler performance: %(rejection_rate).1f%% rej/step, %(mean_nsteps)d steps/it'
                    'mean rel jump distance: %(mean_distance).2f (should be >1), %(frac_far_enough).2f%% (should be >50%%)'
                ) % stepsampler_info

        if ndim == 1:
            pass
        elif ndim == 2 and spearman is not None:
            rho, pval = spearman(p)
            if pval < 0.01 and abs(rho) > 0.75:
                labeltext += ("<br/>   %s between %s and %s: rho=%.2f" % (
                    'positive degeneracy' if rho > 0 else 'negative degeneracy',
                    paramnames[0], paramnames[1], rho))
        elif spearman is not None:
            rho, pval = spearman(p)
            for i, param in enumerate(paramnames):
                for j, param2 in enumerate(paramnames[:i]):
                    if pval[i,j] < 0.01 and abs(rho[i,j]) > 0.99:
                        labeltext += ("<br/>   perfect %s between %s and %s" % (
                            'positive relation' if rho[i,j] > 0 else 'negative relation',
                            param2, param))
                    elif pval[i,j] < 0.01 and abs(rho[i,j]) > 0.75:
                        labeltext += ("<br/>   %s between %s and %s: rho=%.2f" % (
                            'positive degeneracy' if rho[i,j] > 0 else 'negative degeneracy',
                            param2, param, rho[i,j]))

        for i, (_param, fmt) in enumerate(zip(paramnames, paramformats)):
            if nmodes == 1:
                line = [' ' for _ in range(width)]
                for j in np.unique(indices[:,i]):
                    line[j] = '*'
                linestr = ''.join(line)
            else:
                line = [' ' for _ in range(width)]
                for clusterid, j in zip(clusterids, indices[:,i]):
                    if clusterid > 0 and line[j] in (' ', '0'):
                        # set it to correct cluster id
                        line[j] = clusteridstrings[clusterid]
                    elif clusterid == 0 and line[j] == ' ':
                        # empty, so set it although we don't know the cluster id
                        line[j] = '0'
                    # else:
                    #     line[j] = '*'
                linestr = ''.join(line)

            oldlinestr = self.laststatus[i]
            for j, (c, d) in enumerate(zip(linestr, oldlinestr)):
                if c != d:
                    if c == ' ':
                        self.grid[i, j + 2] = 0
                    else:
                        self.grid[i, j + 2] = 1

            self.laststatus[i] = linestr
            # self.grid[i,0].value = param
            self.grid[i, 1] = plo_rounded[i]
            self.grid[i,-1] = phi_rounded[i]

        # self.label.value = labeltext
        self.label = labeltext


"""
    DisguinsingWidgets.py

    Copyright (c) 2016-2018, Masatsuyo Takahashi
"""
import os
import sys
if sys.version_info > (3,):
    import tkinter  as Tk
else:
    import Tkinter  as Tk

class CheckonLabel( Tk.Label ):
    def __init__( self, parent, **kwargs ):
        dir_, _ = os.path.split( __file__ )
        self.cb_image = Tk.PhotoImage( file=os.path.join( dir_, "checkon.gif") )
        Tk.Label.__init__( self, parent, image=self.cb_image, compound=Tk.LEFT, **kwargs )

class CheckoffLabel( Tk.Label ):
    def __init__( self, parent, **kwargs ):
        dir_, _ = os.path.split( __file__ )
        self.cb_image = Tk.PhotoImage( file=os.path.join( dir_, "checkoff.gif") )
        Tk.Label.__init__( self, parent, image=self.cb_image, compound=Tk.LEFT, **kwargs )

if __name__ == '__main__':
    root = Tk.Tk()
    frame = Tk.Frame( root )
    frame.pack( padx=30, pady=10 )
    cb1 = Tk.Checkbutton( frame, text="This is a normal checkbutton." )
    cb1.pack( anchor=Tk.W )

    cbvar = Tk.IntVar()
    cbvar.set( 1 )
    cb2 = Tk.Checkbutton( frame, text="This is a checked and disabled checkbutton.",
                          variable=cbvar, state=Tk.DISABLED)
    cb2.pack( anchor=Tk.W )
    cb3 = CheckonLabel( frame, text="This is a disguising label." )
    cb3.pack( anchor=Tk.W )
    root.mainloop()



"""
    ProportionalDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from molass_legacy.Models.ElutionCurveModels import egha, emga
from DataUtils import get_in_folder

def demo(in_folder, sd):
    import molass_legacy.KekLib.DebugPlot as plt
    from importlib import reload
    import Models.CfsEvalPeaks
    reload(Models.CfsEvalPeaks)
    from .CfsEvalPeaks import get_cfs_evaluated_peaks
    import Models.Proportional
    reload(Models.Proportional)
    from .Proportional import get_proportional_curves

    print(in_folder)
    ecurve = sd.get_xray_curve()
    x = ecurve.x
    y = ecurve.sy

    num_peaks = len(ecurve.peak_info)
    egha_params, emga_params = get_cfs_evaluated_peaks(x, y, num_peaks, debug=True)
    egh_curve_list, emg_curve_list = get_proportional_curves(x, y, egha_params, emga_params)

    in_folder_ = get_in_folder(in_folder)
    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=2, figsize=(12,8))
        fig.suptitle("Proportional Decomposition Demo using %s" % in_folder_, fontsize=20)

        ax1 = axes[0,0]
        ax2 = axes[0,1]
        ax3 = axes[1,0]
        ax4 = axes[1,1]

        for model, func, ax, params in [
                        ("EGHA", egha, ax1, egha_params),
                        ("EMGA", emga, ax2, emga_params),
                        ]:

            ax.set_title("%s Decomposition using CF Space" % model, fontsize=16)

            ax.plot(x, y)

            ty = np.zeros(len(x))
            for p in params:
                cy = func(x, *p)
                ax.plot(x, cy, ":")
                ty += cy

            ax.plot(x, ty, ":", color="red", lw=3)

        for model, ax, curve_list in [
                        ("EGHA", ax3, egh_curve_list),
                        ("EMGA", ax4, emg_curve_list),
                        ]:

            ax.set_title("%s Proportional Decomposition" % model, fontsize=16)

            ax.plot(x, y)

            ty = np.zeros(len(x))
            for cy in curve_list:
                ax.plot(x, cy, ":")
                ty += cy

            ax.plot(x, ty, ":", color="red", lw=3)

        fig.tight_layout()
        fig.subplots_adjust(top=0.9)
        plt.show()



# coding: utf-8
"""
    Peaks.Peak.py

    EmgPeak successor

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from .PeakUtils import recognize_peaks
from .ElutionModels import egh

class Peak:
    def __init__(self, params):
        self.params = params
        self.opt_params = {
                "h" : params[0],
                "mu" : params[1],
                "sigma" : params[2],
                "tau" : params[3],
                }

    def get_model_y(self, x):
        return egh(x, *self.params)

def get_peaks(curve, **kwargs):
    x   = curve.x
    y   = curve.y

    params_list = recognize_peaks(x, y, num_peaks=len(curve.peak_info), debug=True)

    peaks = []
    for params in params_list:
        peaks.append(Peak(params))

    return peaks



# coding: utf-8
"""
    KnownPeakPenaltyNumba.py

    Copyright (c) 2019-2020, SAXS Team, KEK-PF
"""

import numpy as np
try:
    # for numba 1.49 or later
    from numba.core.decorators import jit
except:
    from numba.decorators import jit

@jit(nopython=True)
def compute_known_error_impl(conc_dependence, known_peak_info, P):
    error = 0
    for k, data in enumerate(known_peak_info):
        if data is None:
            continue
        n = k*conc_dependence
        error += np.sum((P[:,n] - data[:,1])**2)
    return error

@jit(nopython=True)
def compute_known_error_grad_impl(conc_dependence, known_peak_info, P):
    error = 0
    for k, data in enumerate(known_peak_info):
        if data is None:
            continue
        n = k*conc_dependence
        error += np.sum(2*(P[:,n] - data[:,1]))
    return error



"""
    Models/RateTheory/EDM.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from matplotlib.widgets import Slider
import molass_legacy.KekLib.DebugPlot as plt
from SecTheory.Edm import edm_func
from molass_legacy.QuickAnalysis.ModeledPeaks import recognize_peaks
from molass_legacy.Models.ElutionCurveModels import egha
from molass_legacy.Models.Tentative import Model
from molass_legacy.Models.ElutionModelUtils import x_from_height_ratio_impl
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.Peaks.ElutionModels import compute_moments

VERY_SMALL_VALUE = 1e-8

def edm_impl(x, t0, u, a, b, e, Dz, cinj):
    return edm_func(x-t0, u, a, b, e, Dz, cinj)

"""
GI      params= [  0.11655683  0.79493376  1.41901069  -0.38470116  0.29435237  0.15273794  1.33639523 ]
GI+ 50  params= [ 50.11653983  0.79492356  1.41896109  -0.38470303  0.29434855  0.1527366   1.33639626 ]
GI+100  params= [100.11653648  0.79494333  1.41899857  -0.38470053  0.29434736  0.15274067  1.33638596 ]
"""

def guess_x0_u(x0, top_x, top_y):
    if top_x < 200:
        x0_ = 0
        u = 0.7         # ok for sample_data
        # u = 1.0         # ok for HasA
    elif top_x < 300:
        x0_ = 0 if x0 < 200 else x0/2
        u = 0.7
        # u = 0.5
    else:
        x0_ = top_x/3
        u = 0.5
    return x0_, u

GUESS_PARAMS = np.array([
    [-0.403, 0.00278, -8.7e-05, -6.21e-06],
    [1.3, -0.00749, 0.000712, -1.02e-05],
    [1.42, 0.00207, -0.000579, 2.38e-05],
    [-12.1, 0.063, -0.00155, -0.000146],
    [0.477, 0.00916, -0.00199, 1.79e-05],
    [0.496, -0.00305, 0.000205, 1.31e-06],
    [-2.18, 0.0203, -0.00129, 4.19e-06],
    ])

def guess_init_params(moments):
    return [GUESS_PARAMS[k,0] + np.sum(GUESS_PARAMS[k,1:] * np.array(moments)) for k in range(7)]

def guess(x, y, moments=None, init_params=None, debug=False):
    if moments is None:
        if init_params is None:
            j = np.argmax(y)
            top_y = y[j]
            top_x = x[j]
            print("top_y=", top_y)
            x0, u = guess_x0_u(x[0], top_x, top_y)
            cinj_init = min(1.0, 1.32 * top_y/0.291)
            init_params = [x0, u, 1.5, -3.0, 0.4, 0.06, cinj_init]
    else:
        init_params = guess_init_params(moments)

    print("init_params=", init_params)

    if debug:
        slider_specs = [    ("t0", -50, 200, init_params[0]),
                            ("u", 0, 2, init_params[1]),
                            ("a", 0, 2, init_params[3]),
                            ("b", -4, 1, init_params[3]),
                            ("e", 0, 2, init_params[4]),
                            ("Dz", 0, 1, init_params[5]),
                            ("cinj", 0, 3, init_params[6]),
                            # ("tinj", 0, 10, init_params[7]),
                            ]

        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(x, y)

            edm_line, = ax.plot(x, edm_impl(x, *init_params))

            def slider_update(k, val):
                # print([k], "slider_update", val)
                init_params[k] = val
                y_ = edm_impl(x, *init_params)
                edm_line.set_data(x, y_)
                dp = plt.get_dp()
                dp.draw()

            slider_axes = []
            sliders = []
            for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
                ax_ = fig.add_axes([0.15, 0.8 - 0.08*k, 0.25, 0.03])
                slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
                slider.on_changed(lambda val, k_=k: slider_update(k_, val))
                slider_axes.append(ax_)
                sliders.append(slider)

            fig.tight_layout()
            plt.show()

    def objective(p):
        y_ = edm_impl(x, *p)
        return np.sum((y_ - y)**2)

    ret = minimize(objective, init_params)

    if debug:
        M0 = np.sum(y)
        print("ret.x=", ret.x, M0/30)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(x, y)
            ax.plot(x, edm_impl(x, *ret.x))
            fig.tight_layout()
            plt.show()

    return ret.x

def guess_multiple(x, y, num_peaks, fh=None, debug=False):

    M = compute_moments(x, y)
    params = guess(x, y, moments=M, debug=True)

    if fh is not None:
        fh.write(",".join(["%.3g" % v for v in [*M, *params]]) + "\n")

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.plot(x, y)
        fig.tight_layout()
        plt.show()

class EDM(Model):
    def __init__(self, **kwargs):
        super(EDM, self).__init__(edm_impl, **kwargs)

    def get_name(self):
        return "EDM"

    def is_traditional(self):
        return False

    def eval(self, params=None, x=None):
        return self.func(x, *params)

    def x_from_height_ratio(self, ecurve, ratio, params):
        return x_from_height_ratio_impl(edm_impl, ecurve, ratio, *params, needs_ymax=True, full_params=True)

    def get_params_string(self, params):
        return 't0=%g, u=%g, a=%g, b=%g, e=%g, Dz=%g, cinj=%g' % tuple(params)



"""
    ExcelManager.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import os
from time import sleep

class ExcelManager:
    def __init__(self):
        pass

    def manage_loop(self, job_queue, log_folder):
        from molass_legacy.KekLib.ExcelCOM import CoInitialize, CoUninitialize, ExcelComClient
        from molass_legacy.KekLib.ChangeableLogger import Logger
        from molass_legacy._MOLASS.SerialSettings import initialize_settings
        print('manage_loop')
        self.stop = False
        self.interval = 1
        self.job_queue = job_queue
        self.logger = Logger(log_folder+'/manager.log')
        self.logger.info("manager start in process %d", os.getppid())
        initialize_settings(warn_on_fail=False)     # suppress the warning message
        self.logger.info("initialize_settings() done")
        CoInitialize()
        self.excel_client = ExcelComClient()
        self.logger.info("excel_client created with pid=%d", self.excel_client.get_pid())

        while not self.stop:
            print('managing')
            if self.job_queue.empty():
                sleep(self.interval)
            else:
                info = self.job_queue.get()
                if info.code == 'stop':
                    break
                self.execute(info)

        self.close()
        CoUninitialize()

    def close(self):
        # self.excel_client.quit()
        pid = self.excel_client.get_pid()
        del self.excel_client
        self.excel_client = None
        self.logger.info("excel_client with pid=%d destroyed.", pid)

    def execute(self, info):
        self.logger.info('executing %s', info.code)

        try:
            action = self.get_action(info)
            if action is None:
                raise RuntimeError(info)
            else:
                action(info.args)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            self.logger.error(str(etb))

    def get_action(self, info):
        action = self.__getattribute__(info.code)
        return action

    def guinier_book(self, args):
        from molass_legacy.Reports.GuinierExcelFormatter import add_guinier_annonations
        self.logger.info('gunier_book: %s', str(args))
        add_guinier_annonations(self.excel_client, args, self.logger)

    def range_extrapolation_book(self, args):
        from molass_legacy.Reports.ZeroExExcelFormatter import add_result_format_setting
        self.logger.info('range_extrapolation_book: %s', str(args))
        add_result_format_setting(self.excel_client, args, self.logger)

    def range_datgnom_book(self, args):
        from molass_legacy.Reports.DatgnomExcelFormatter import add_datgnom_format_setting
        self.logger.info('range_datgnom_book: %s', str(args))
        add_datgnom_format_setting(self.excel_client, args, self.logger)

    def range_overlay_book(self, args):
        from molass_legacy.Reports.ZeroExExcelFormatter import add_overlay_format_setting
        self.logger.info('range_overlay_book: %s', str(args))
        add_overlay_format_setting(self.excel_client, args, self.logger)

    def summary_book(self, args):
        from molass_legacy.Reports.SummaryExcelFormatter import add_summary_format_setting
        self.logger.info('summary_book: %s', str(args))
        add_summary_format_setting(self.excel_client, args, self.logger)

    def merge_excel_books(self, args):
        from molass_legacy.SerialAnalyzer.StageSummary import merge_books_impl
        self.logger.info('merge_excel_books: %s', str(args))
        merge_books_impl(self.excel_client, args, self.logger)



# coding: utf-8
"""

    ファイル名：   AnalyzerUtil.py

    処理内容：

        連続測定データ全体解析起動 GUI の範囲設定

    Copyright (c) 2016-2020, SAXS Team, KEK-PF

"""
import os
from molass_legacy.KekLib.OurTkinter             import Tk
from molass_legacy._MOLASS.SerialSettings         import get_setting
from molass_legacy.KekLib.BasicUtils             import mkdirs_with_retry, get_caller_module
from molass_legacy._MOLASS.Version                import get_version_string
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting
from molass_legacy.DataStructure.AnalysisRangeInfo      import AnalysisRangeInfo

def make_guinier_result_stamp( self ):

    input_smoothing         = get_setting( 'input_smoothing' )
    if input_smoothing == 1:
        params = 'Averaging ' + str( get_setting( 'num_curves_averaged' ) ) + 'curves\n'
        smoothing_description  = 'Smoothing Method on Elution Axis,' + params
    else:
        smoothing_description  = 'No Smoothing on Elution Axis\n'

    stamp = (   'Serial Analyzer Version,' + get_version_string() + '\n'
                + 'Intensity Data Folder,' + self.data_folder + '\n'
                + 'Absorbance Data Folder,' + self.conc_folder + '\n'
                + smoothing_description
                )
    return stamp

def exists_guinier_analysis_result( self, create=True ):
    if self.guinier_folder is None: return False
    if self.stamp_file is None: return False

    if not os.path.exists( self.guinier_folder ):
        if create:
            mkdirs_with_retry( self.guinier_folder )
        return False

    if not os.path.exists( self.stamp_file ):
        return False

    try:
        fh = open( self.stamp_file )
        stamp = fh.read()
        fh.close()
    except:
        stamp = ''

    return stamp == make_guinier_result_stamp( self )

def compute_conc_factor_util():
    return get_setting( 'path_length' ) / get_setting( 'extinction' )

def get_init_ranges( just_ranges=None ):
    ranges = []

    analysis_range_info = get_setting( 'analysis_range_info' )
    if analysis_range_info is None:
        manual_range_info = get_setting( 'manual_range_info' )
        if manual_range_info is not None:
            just_ranges = manual_range_info
        for range_ in just_ranges:
            # print('get_init_ranges(1): range_=', range_)
            ranges.append( [ 1, tuple( range_ ) ] )
    else:
        for range_ in analysis_range_info.get_ranges():
            # print('get_init_ranges(2): range_=', range_)
            top_x = int(range_[0].top_x + 0.5)
            ranges.append( [ 1, (range_[1][0], top_x, range_[-1][1]) ] )

    return ranges

def set_analysis_ranges( ranges ):
    set_setting( 'analysis_range_info', AnalysisRangeInfo(ranges) )

def get_analysis_ranges():
    analysis_range_info = get_setting( 'analysis_range_info' )
    assert analysis_range_info is not None
    return analysis_range_info.get_ranges()



# coding: utf-8
"""
    SecTools.CorMap.PairwiseMap.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from MatrixData import simple_plot_3d

class PairwiseMap(Dialog):
    def __init__(self, parent, M, j_pair, qv, datcmp_data):
        self.M = M
        self.j_pair = j_pair
        self.j_pair_ = tuple([j+1 for j in j_pair])
        self.R = np.corrcoef(M[:,j_pair])
        self.qv = qv
        self.datcmp_data = datcmp_data
        self.pairwise_point = None
        self.pairwise_lines = None
        Dialog.__init__(self, parent, "Pairwise Map", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig = plt.figure(figsize=(18,6))
        self.fig = fig
        ax1 = fig.add_subplot(131, projection="3d")
        ax2 = fig.add_subplot(132)
        ax3 = fig.add_subplot(133)

        ax1.set_title("Pair Frames in 3D", fontsize=16)
        ax2.set_title("Pair Correlation between Frames(%d, %d)" % self.j_pair, fontsize=16)
        ax3.set_title("P-value in Schilling Distribution (n=%d)" % len(self.qv), fontsize=16)
        self.axes = ax1, ax2, ax3

        self.draw_pairframes_3d(ax1)
        self.draw_pairwise_correlation(ax2)
        self.draw_pvalue_in_schilling(ax3)

        fig.tight_layout()
        fig.subplots_adjust(bottom=0.1)

        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def draw_pairframes_3d(self, ax):
        ax.set_xlabel("$q(\AA^{-1})$")
        ax.set_ylabel("Frame No.")
        ax.set_zlabel("Intensity")
        simple_plot_3d(ax, self.M, x=self.qv)
        x = self.qv
        for j in self.j_pair:
            y = np.ones(len(x)) * j
            z = self.M[:,j]
            ax.plot(x, y, z, label="Frame No. %d" % j)
        ax.legend()

    def draw_pairwise_correlation(self, ax):
        ax.set_xlabel(r"$q(\AA^{-1})$")
        ax.set_ylabel(r"$q(\AA^{-1})$")

        f, t = self.qv[[0, -1]]
        ax.imshow(self.R, extent=(f, t, t, f))

    def get_datcmp_rec(self):
        datcmp_key = "%d,%d" % self.j_pair_
        return self.datcmp_data.get(datcmp_key)

    def draw_pvalue_in_schilling(self, ax):
        from .SchillingDist import schilling_pdf
        ax.set_xlabel("Longest Run")
        ax.set_ylabel("Density")
        n = len(self.qv)
        x = np.linspace(2, 30, 100)
        y = schilling_pdf(n, 0.5, x)
        ax.plot(x, y)
        length, p_val, p_adj = self.get_datcmp_rec()
        x_ge = x >= length
        x_ = x[x_ge]
        y_ = y[x_ge]
        ax.fill_between(x_, y_, color="red", alpha=0.3)
        ymin, ymax = ax.get_ylim()
        def draw_text(tx, tyw, text):
            ty = ymin * (1 - tyw) + ymax * tyw
            ax.text(tx, ty, text, fontsize=16)

        draw_text(15, 0.8, "Longest Run = %.3g" % length)
        draw_text(15, 0.7, "P-value     = %.3g" % p_val)
        draw_text(15, 0.6, "P-adjusted  = %.3g" % p_adj)

    def on_figure_click(self, event):
        if event.button != 1:
            return

        ax2 = self.axes[1]
        if event.inaxes != ax2:
            return

        self.draw_pointed_data(event)
        self.mpl_canvas.draw()

    def draw_pointed_data(self, event):
        x = event.xdata
        y = event.ydata
        min_j = 0
        max_j = len(self.qv) - 1
        def get_index(x):
            j = bisect_right(self.qv, x)
            return max(min_j, min(max_j, j))

        i, j = tuple(sorted([get_index(v) for v in [x, y]]))
        xyz_list = []
        for k in (i, j):
            x_ = self.qv[[k,k]]
            y_ = self.j_pair
            z_ = self.M[k,y_]
            xyz_list.append((x_, y_, z_))
        if self.pairwise_point is None:
            ax1, ax2 = self.axes[0:2]
            self.pairwise_point, = ax2.plot(x, y, 'o', color="cyan")
            lines = []
            for k, (x_, y_, z_), c in zip((i, j), xyz_list, ["red", "yellow"]):
                line, = ax1.plot(x_, y_, z_, "-o", color=c, label="points at %.3g" % self.qv[k])
                lines.append(line)
            self.ax1_legend = ax1.legend()
            self.pairwise_lines = lines
        else:
            self.pairwise_point.set_data(x, y)
            for line, (x_, y_, z_) in zip(self.pairwise_lines, xyz_list):
                line.set_xdata(x_)
                line.set_ydata(y_)
                line.set_3d_properties(z_)
            for text, q in zip(self.ax1_legend.get_texts()[2:4], self.qv[[i,j]]):
                text.set_text("points at %.3g" % q)

        self.mpl_canvas.draw()



"""
    OurGeomdl.py

    modified from geomdl/visualization/VisMPL.VisSurface

"""

from geomdl import vis
from geomdl.visualization.VisMPL import VisConfig
import numpy as np
import logging
import matplotlib as mpl
import matplotlib.tri as mpltri
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from matplotlib import animation

class VisSurface(vis.VisAbstract):
    """ Matplotlib visualization module for surfaces.

    Wireframe plot for the control points and triangulated plot (using ``plot_trisurf``) for the surface points.
    The surface is triangulated externally using :py:func:`.utilities.make_triangle_mesh()` function.
    """
    def __init__(self, config=VisConfig(), **kwargs):
        super(VisSurface, self).__init__(config, **kwargs)
        self._module_config['ctrlpts'] = "quads"
        self._module_config['evalpts'] = "triangles"
        self.fig = kwargs.pop('fig', None)
        self.ax = kwargs.pop('ax', None)
        alpha = kwargs.pop('alpha', None)
        if alpha is not None:
            self.vconf.alpha = alpha
        self.logger = logging.getLogger( __name__ )
        self.logger.info( "using modified version of geomdl.visualization.VisMPL.VisSurface. alpha=%g" % self.vconf.alpha )

    def animate(self, **kwargs):
        """ Animates the surface.

        This function only animates the triangulated surface. There will be no other elements, such as control points
        grid or bounding box.

        Keyword arguments:
            * ``colormap``: applies colormap to the surface

        Colormaps are a visualization feature of Matplotlib. They can be used for several types of surface plots via
        the following import statement: ``from matplotlib import cm``

        The following link displays the list of Matplolib colormaps and some examples on colormaps:
        https://matplotlib.org/tutorials/colors/colormaps.html
        """
        # Calling parent render function
        super(VisSurface, self).render(**kwargs)

        # Colormaps
        surf_cmaps = kwargs.get('colormap', None)

        # Initialize variables
        tri_idxs = []
        vert_coords = []
        trisurf_params = []
        frames = []
        frames_tris = []
        num_vertices = 0

        # Start plotting of the surface and the control points grid
        fig = plt.figure(figsize=self.vconf.figure_size, dpi=self.vconf.figure_dpi)
        ax = Axes3D(fig)

        # Start plotting
        surf_count = 0
        for plot in self._plots:
            # Plot evaluated points
            if plot['type'] == 'evalpts' and self.vconf.display_evalpts:
                # Use internal triangulation algorithm instead of Qhull (MPL default)
                verts = plot['ptsarr'][0]
                tris = plot['ptsarr'][1]
                # Extract zero-indexed vertex number list
                tri_idxs += [[ti + num_vertices for ti in tri.data] for tri in tris]
                # Extract vertex coordinates
                vert_coords += [vert.data for vert in verts]
                # Update number of vertices
                num_vertices = len(vert_coords)

                # Determine the color or the colormap of the triangulated plot
                params = {}
                if surf_cmaps:
                    try:
                        params['cmap'] = surf_cmaps[surf_count]
                        surf_count += 1
                    except IndexError:
                        params['color'] = plot['color']
                else:
                    params['color'] = plot['color']
                trisurf_params += [params for _ in range(len(tris))]

        # Pre-processing for the animation
        pts = np.array(vert_coords, dtype=self.vconf.dtype)

        # Create the frames (Artists)
        for tidx, pidx in zip(tri_idxs, trisurf_params):
            frames_tris.append(tidx)
            # Create MPL Triangulation object
            triangulation = mpltri.Triangulation(pts[:, 0], pts[:, 1], triangles=frames_tris)
            # Use custom Triangulation object and the choice of color/colormap to plot the surface
            p3df = ax.plot_trisurf(triangulation, pts[:, 2], alpha=self.vconf.alpha, **pidx)
            # Add to frames list
            frames.append([p3df])

        # Create MPL ArtistAnimation
        ani = animation.ArtistAnimation(fig, frames, interval=100, blit=True, repeat_delay=1000)

        # Remove axes
        if not self.vconf.display_axes:
            plt.axis('off')

        # Set axes equal
        if self.vconf.axes_equal:
            self.vconf.set_axes_equal(ax)

        # Axis labels
        if self.vconf.display_labels:
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')

        # Process keyword arguments
        fig_filename = kwargs.get('fig_save_as', None)
        fig_display = kwargs.get('display_plot', True)

        # Display the plot
        if fig_display:
            plt.show()
        else:
            fig_filename = self.vconf.figure_image_filename if fig_filename is None else fig_filename

        # Save the figure
        self.vconf.save_figure_as(fig, fig_filename)

        # Return the figure object
        return fig

    def render(self, **kwargs):
        """ Plots the surface and the control points grid.

        Keyword arguments:
            * ``colormap``: applies colormap to the surface

        Colormaps are a visualization feature of Matplotlib. They can be used for several types of surface plots via
        the following import statement: ``from matplotlib import cm``

        The following link displays the list of Matplolib colormaps and some examples on colormaps:
        https://matplotlib.org/tutorials/colors/colormaps.html
        """
        # Calling parent function
        super(VisSurface, self).render(**kwargs)

        # Colormaps
        surf_cmaps = kwargs.get('colormap', None)

        # Initialize variables
        legend_proxy = []
        legend_names = []

        # Start plotting of the surface and the control points grid
        if self.fig is None:
            fig = plt.figure(figsize=self.vconf.figure_size, dpi=self.vconf.figure_dpi)
            ax = Axes3D(fig)
        else:
            fig = self.fig
            ax = self.ax
            assert ax is not None

        surf_count = 0
        # Start plotting
        for plot in self._plots:
            # Plot control points
            if plot['type'] == 'ctrlpts' and self.vconf.display_ctrlpts:
                vertices = [v.data for v in plot['ptsarr'][0]]
                faces = [q.data for q in plot['ptsarr'][1]]
                for q in faces:
                    el = np.array([vertices[i] for i in q], dtype=self.vconf.dtype)
                    el[:, 2] += self._ctrlpts_offset
                    pc3d = Poly3DCollection([el], alpha=0.0, edgecolors=plot['color'], linewidths=1.0, linestyles='-.')
                    pc3d.set_facecolor(None)
                    ax.add_collection3d(pc3d)
                pts = np.array(vertices, dtype=self.vconf.dtype)
                pts[:, 2] += self._ctrlpts_offset
                ax.scatter(pts[:, 0], pts[:, 1], pts[:, 2], color=plot['color'], linestyle='-.', marker='o')
                plot_proxy = mpl.lines.Line2D([0], [0], linestyle='-.', color=plot['color'], marker='o')
                legend_proxy.append(plot_proxy)
                legend_names.append(plot['name'])

            # Plot evaluated points
            if plot['type'] == 'evalpts' and self.vconf.display_evalpts:
                # Use internal triangulation algorithm instead of Qhull (MPL default)
                verts = plot['ptsarr'][0]
                tris = plot['ptsarr'][1]
                # Extract zero-indexed vertex number list
                tri_idxs = [tri.data for tri in tris]
                # Extract vertex coordinates
                vert_coords = [vert.data for vert in verts]
                pts = np.array(vert_coords, dtype=self.vconf.dtype)

                # Determine the color or the colormap of the triangulated plot
                trisurf_params = {}
                if surf_cmaps:
                    try:
                        trisurf_params['cmap'] = surf_cmaps[surf_count]
                        surf_count += 1
                    except IndexError:
                        trisurf_params['color'] = plot['color']
                else:
                    trisurf_params['color'] = plot['color']

                # Create MPL Triangulation object
                if pts.size != 0:
                    triangulation = mpltri.Triangulation(pts[:, 0], pts[:, 1], triangles=tri_idxs)
                    # Use custom Triangulation object and the choice of color/colormap to plot the surface
                    ax.plot_trisurf(triangulation, pts[:, 2], alpha=self.vconf.alpha, **trisurf_params)
                    # Add to legend
                    plot_proxy = mpl.lines.Line2D([0], [0], linestyle='none', color=plot['color'], marker='^')
                    legend_proxy.append(plot_proxy)
                    legend_names.append(plot['name'])

            # Plot bounding box
            if plot['type'] == 'bbox' and self.vconf.display_bbox:
                pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
                ax.plot(pts[:, 0], pts[:, 1], pts[:, 2], color=plot['color'], linestyle='--')
                plot_proxy = mpl.lines.Line2D([0], [0], linestyle='--', color=plot['color'])
                legend_proxy.append(plot_proxy)
                legend_names.append(plot['name'])

            # Plot trim curves
            if self.vconf.display_trims:
                if plot['type'] == 'trimcurve':
                    pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
                    ax.scatter(pts[:, 0], pts[:, 1], pts[:, 2], color=plot['color'], marker='o',
                               s=self.vconf.trim_size, depthshade=False)
                    plot_proxy = mpl.lines.Line2D([0], [0], linestyle='none', color=plot['color'], marker='o')
                    legend_proxy.append(plot_proxy)
                    legend_names.append(plot['name'])

            # Plot extras
            if plot['type'] == 'extras':
                pts = np.array(plot['ptsarr'], dtype=self.vconf.dtype)
                ax.plot(pts[:, 0], pts[:, 1], pts[:, 2],
                        color=plot['color'][0], linestyle='-', linewidth=plot['color'][1])
                plot_proxy = mpl.lines.Line2D([0], [0], linestyle='-', color=plot['color'][0])
                legend_proxy.append(plot_proxy)
                legend_names.append(plot['name'])

        # Add legend to 3D plot, @ref: https://stackoverflow.com/a/20505720
        if self.vconf.display_legend:
            ax.legend(legend_proxy, legend_names, numpoints=1)

        # Remove axes
        if not self.vconf.display_axes:
            plt.axis('off')

        # Set axes equal
        if self.vconf.axes_equal:
            self.vconf.set_axes_equal(ax)

        # Axis labels
        if self.vconf.display_labels:
            ax.set_xlabel('x')
            ax.set_ylabel('y')
            ax.set_zlabel('z')

        # Process keyword arguments
        fig_filename = kwargs.get('fig_save_as', None)
        fig_display = kwargs.get('display_plot', True)

        # Display the plot
        if fig_display:
            plt.show()
        else:
            fig_filename = self.vconf.figure_image_filename if fig_filename is None else fig_filename

        # Save the figure
        self.vconf.save_figure_as(fig, fig_filename)

        # Return the figure object
        return fig



# coding: utf-8
"""
    Baseline.LpmProxy.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np

class LpmProxy:
    def __init__(self, data, **kwargs):
        thru_mode = kwargs.pop('thru_mode', False)
        self.qvector = kwargs.pop('q')
        self.mapper = kwargs.pop('mapper')
        if thru_mode:
            from LPM import LPM_3d
            lpm = LPM_3d(data, **kwargs)
            self.data = lpm.data
        else:
            self.data = self.correct_old_style(data, **kwargs)

    def correct_old_style(self, data, **kwargs):
        from ScatteringBaseCorrector import ScatteringBaseCorrector
        jsize = data.shape[1]
        qvector = self.qvector
        jvector = np.arange(jsize)
        assert qvector is not None

        evector = np.zeros(len(qvector))
        intensity_array = np.array([np.vstack([qvector, data[:,j], evector]).T for j in range(jsize)])
        print('intensity_array.shape=', intensity_array.shape)

        # affine_info = [jvector, np.zeros(jsize), 0]
        affine_info = self.mapper.get_affine_info()

        inty_curve_y = kwargs.pop('ecurve_y')
        corrector = ScatteringBaseCorrector(jvector, qvector, intensity_array,
                        curve=self.mapper.sd_xray_curve
                        affine_info=affine_info, inty_curve_y=inty_curve_y,
                        )

        corrector.correct_all_q_planes()
        corrected = intensity_array[:,:,1].T
        print('corrected.shape=', corrected.shape)

        return corrected



"""
    SecTheory.ExclCurveUtils.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import UnivariateSpline

def safely_create_spline_from_points(points):
    x = points[:,0]
    y = points[:,1]

    diff_x = np.diff(x)
    strictly_increasing = np.concatenate([[0], np.where(diff_x > 1e-6)[0] + 1])

    x_ = x[strictly_increasing]
    y_ = y[strictly_increasing]

    try:
        spline = UnivariateSpline(x_, y_, s=0)
    except:
        from molass_legacy.KekLib.ExceptionTracebacker import log_exception
        import molass_legacy.KekLib.DebugPlot as plt
        log_exception(None, "UnivariateSpline: ")

        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("safely_create_spline_from_points: debug")
            ax.plot(x_, y_, "o")
            fig.tight_layout()
            plt.show()

    return spline

if __name__ == '__main__':
    import os
    import sys

    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)



"""
    adapted from
        Radial distribution function
        https://pythoninchemistry.org/sim_and_scat/calculating_scattering/rdfs.html
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from pylj import md, sample

def md_simulation(number_of_particles, temperature, 
                  box_length, number_of_steps, 
                  sample_frequency):
    """
    Runs a molecular dynamics simulation in using the pylj 
    molecular dynamics engine.
    
    Parameters
    ----------
    number_of_particles: int
        The number of particles in the simulation
    temperature: float
        The temperature for the initialisation and 
        thermostating
    box_length: float
        The length of the simulation square
    number_of_steps: int
        The number of molecular dynamics steps to run
    sample_frequency: 
        How regularly the visualisation should be updated
        
    Returns
    -------
    pylj.util.System
        The complete system information from pylj
    """
    # %matplotlib notebook
    system = md.initialise(number_of_particles, temperature, 
                           box_length, 'square')
    xpos = system.particles["xposition"].copy()
    ypos = system.particles["yposition"].copy()
    x = []
    y = []
    average_rdf = []
    r = []
    system.time = 0
    for i in range(0, number_of_steps):
        system.integrate(md.velocity_verlet)
        system.md_sample()
        system.heat_bath(temperature)
        system.time += system.timestep_length
        system.step += 1
        if system.step % sample_frequency == 0:
            # sample_system.update(system)
            modified_update(system, x, y, average_rdf, r)

    make_animation(system, xpos, ypos, x, y, average_rdf, r)

def modified_update(system, x, y, average_rdf, r):
    x3 = system.particles["xposition"].copy()
    y3 = system.particles["yposition"].copy()
    x.append(x3)
    y.append(y3)
    hist, bin_edges = np.histogram(
        system.distances, bins=np.linspace(0, system.box_length / 2 + 0.5e-10, 100)
    )
    gr = hist / (
        system.number_of_particles
        * (system.number_of_particles / system.box_length ** 2)
        * np.pi
        * (bin_edges[:-1] + 0.5e-10 / 2.0)
        * 0.5
    )
    average_rdf.append(gr)
    x = bin_edges[:-1] + 0.5e-10 / 2
    r.append(x)

def make_animation(system, xpos, ypos, x, y, average_rdf, r, scale = 1):
    gr = np.average(average_rdf, axis=0)
    ar = np.average(r, axis=0)

    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,6))
    fig.suptitle("MD Simulation Demo using pylj @ https://pythoninchemistry.org/", fontsize=20)
    ax1.set_title("Particle Positions", fontsize=16)
    ax2.set_title("Snapshot RDF", fontsize=16)
    ax3.set_title("Averaged RDF", fontsize=16)

    mk = 6.00555e-8 / (system.box_length - 2.2727e-10) - 1e-10 / scale
    art1, = ax1.plot(xpos, ypos, "o", markersize=mk, markeredgecolor="black", color="#34a5daff")
    ax1.set_xlim([0, system.box_length])
    ax1.set_ylim([0, system.box_length])
    ax1.set_xticks([])
    ax1.set_yticks([])

    for ax in [ax2, ax3]:
        ax.set_xlim([0, system.box_length / 2])
        # ax.set_yticks([])
        ax.set_ylabel("RDF", fontsize=16)
        ax.set_xlabel("r/m", fontsize=16)

    art2, = ax2.plot([0], color="#34a5daff")
    ax3.plot(ar, gr, color="#34a5daff")
    ymin, ymax = ax3.get_ylim()
    ymax *= 1.2
    for ax in [ax2, ax3]:
        ax.set_ylim(ymin, ymax)

    def init():
        return art1, art2

    def update(i):
        art1.set_data(x[i], y[i])
        gr = average_rdf[i]
        art2.set_data(r[i], gr)
        return art1, art2

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)

    num_frames = len(x)
    print("num_frames=", num_frames)
    ani = FuncAnimation(fig, update, frames=num_frames,
                        init_func=init, blit=True)
    plt.show()

    # ani.save('pylj-md_simulation.mp4', writer="ffmpeg")
    ani.save('pylj-md_simulation.gif')

if __name__ == '__main__':
    md_simulation(20, 300, 20, 2000, 25)



"""

    GuiMain.py

    Copyright (c) 2016-2025, SAXS Team, KEK-PF

"""
import sys
import re
import os
import time
import warnings
from molass_legacy.KekLib.MultiMonitor import get_selected_monitor
from molass_legacy.KekLib.EnvironCheck import executables_check
from molass_legacy.KekLib.OurMatplotlib import mpl_1_5_backward_compatible_init, mpl_font_init
mpl_1_5_backward_compatible_init()
mpl_font_init()

from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker, log_exception
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry, is_almost_empty_dir
from molass_legacy.KekLib.ChangeableLogger import Logger
from molass_legacy.KekLib.OurTkinter import Tk, ttk, FileDialog, Font, is_empty_val
from molass_legacy.KekLib.TkUtils import adjusted_geometry, PositionSynchronizer
from molass_legacy.KekLib.TkSupplements import SlimButton, BlinkingFrame
from molass_legacy.KekLib.TkSizableWidgets import SizableLabel, SizableEntry, BlinkingLabel
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry, FileEntry
try:
    import molass_legacy.KekLib.CustomMessageBox as MessageBox
except:
    import molass_legacy.KekLib.OurMessageBox as MessageBox
print( 'import molass_legacy.KekLib.CustomMessageBox ok' )
from molass_legacy.KekLib.RecentFolders import RecentFolders
from molass_legacy.Menus.GuiSettings import GuiSettingsMenu
from molass_legacy.Menus.GuiSecTools import GuiSecToolsMenu
from molass_legacy.Menus.GuiMctTools import GuiMctToolsMenu
from molass_legacy.Menus.GuiDenssTools import GuiDenssToolsMenu
from molass_legacy.Menus.GuiTutorials import GuiTutorialsMenu
from molass_legacy.Menus.GuiReferences import GuiReferencesMenu
from molass_legacy.Menus.GuiDevelopment import GuiDevelopmentMenu

from molass_legacy.SerialAnalyzer.SerialDataLoader import SerialDataLoader
from molass_legacy.SerialAnalyzer.SerialDataUtils import get_uv_filename, get_mtd_filename
from molass_legacy.DataStructure.MeasuredData import MeasuredData
from molass_legacy.Trimming.OutlineFigure import OutlineFigure
from molass_legacy.SerialAnalyzer.AbnormalityCheck import exclude_abnormality
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.QuickAnalysis.Analyzer import Analyzer
from molass_legacy._MOLASS.Version import get_version_string, is_developing_version
from molass_legacy.Rank.RankView import RankViewMenu
from molass_legacy.GuiParts.UvFolderEntry import UvFolderEntry
from molass_legacy.Menus.V2Menu import V2Menu
from molass_legacy._MOLASS.SerialSettings import ( initialize_settings,
                                    get_setting, set_setting, save_settings,
                                    clear_temporary_settings, restore_default_setting,
                                    get_settings_folder )
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting, set_dev_setting
import molass_legacy.AutorgKek.Settings

analysis_name_re    = re.compile( r'^(.*?)(\d+)(.*)$' )
distance_symbol_re  = re.compile( r'^(\S+)\s' )

label_bg_color = 'gray40'
label_fg_color = 'white'
label_gb_color_var = 'white smoke'
relief_var = Tk.GROOVE
radio_button_text_length = 20

angular_unit_guide = "; this unit is used only for display purpose."

class GuiMain( Tk.Toplevel ):
    def __init__(self, parent, pid=None, debug=0):
        from molass_legacy._MOLASS.Debug import DebugInfo
        dpi_aware = False
        self.debug_info = DebugInfo(debug=debug)
        self.clear_test_mode()

        self.parent_pid = pid
        self.cleaner = None

        if dpi_aware:
            from molass_legacy.KekLib.DpiAware import set_dpi_aware
            set_dpi_aware()
            monitor = get_selected_monitor()
            self.high_dpi = monitor.width >= 2560
            if self.high_dpi:
                set_dev_setting('hiresolution', True)
        else:
            self.high_dpi = False

        self.testing = False
        if not executables_check(parent):
            parent.after(100, parent.quit)
            return

        initialize_settings()

        self.single_instance = None     # this is required in case of exiting within the following check
        from molass_legacy._MOLASS.AppSingleInstance import single_instance_check
        app_lock_path = get_settings_folder() + '\\lock'
        self.single_instance = single_instance_check(app_lock_path)

        self.default_font = Font.nametofont("TkDefaultFont")
        self.hiresolution   = get_dev_setting( 'hiresolution' )

        if self.hiresolution:
            if self.high_dpi:
                fixed_fontsize = 10
                self.slimbutton = True
            else:
                fixed_fontsize = 16
                self.slimbutton = False
            self.default_font.configure(size=fixed_fontsize)
            parent.option_add("*Font", self.default_font)
        else:
            self.slimbutton = True
            fixed_fontsize  = 9

        self.fixed_font = Font.Font( family="Courier", size=fixed_fontsize )

        warnings.filterwarnings("ignore")

        parent.wm_title( get_version_string() + " pid:%d" % os.getpid() )
        parent.report_callback_exception = self.report_callback_exception

        if '-i' in sys.argv:
            from molass_legacy._MOLASS.SerialSettings   import clear_settings
            clear_settings()

        self.parent = parent
        self.tmp_logger = Logger()  # this logger logs into string stream
        clear_temporary_settings()
        self.serial_data = None
        self.measured_data = None

        self.loader     = SerialDataLoader(dialog=self)
        self.analyzer   = Analyzer( self, self.loader )
        self.intial_geometry = None
        self.use_mtd_conc = False
        self.is_executing   = False
        self.dataset_is_ready = False
        self.preproc_dialog = None
        self.keeping_setting_items = None

        Tk.Toplevel.__init__( self, parent )
        self.withdraw()

        if get_setting('enable_debug_plot'):
            from molass_legacy.KekLib.DebugPlot import set_plot_env
            set_plot_env(self)

        self.build_window()
        if '--close_immediately' in sys.argv:
            print( 'closing immediately' )
            self.quit( immediately=True )

        self.after(100, self.init_proc)
        self.pos_sync = PositionSynchronizer(self, self.parent)

    def __del__(self):
        if self.single_instance is not None:
            self.single_instance.clean_up()

    def set_test_mode(self, tester_info):
        self.testing = True
        self.tester_info = tester_info
        set_setting("test_pattern", tester_info.test_pattern)

    def clear_test_mode(self):
        self.testing = False
        self.tester_info = None
        set_setting("test_pattern", None)

    def init_proc(self):
        # initial proc done after GuiMain.__init__
        for k in range(2):
            try:
                self.check_environment()
                self.make_it_more_clever()
                break
            except:
                # ocurred once when D&D in a not ready state
                pass
            time.sleep(1)

        self.after(1000, self.update_menu_states)   # to ensure DENSS Tools menu states update

    def check_environment( self ):
        from molass_legacy.Env.EnvInfo import EnvInfo, set_global_env_info
        env_info = EnvInfo()
        env_info.show_and_log_if_unavailable(self, self.tmp_logger)
        self.env_info = env_info
        set_global_env_info(env_info)

    def make_it_more_clever(self):
        from molass_legacy.SerialAnalyzer.MakeItMoreClever import increase_menu_availability
        increase_menu_availability(self)

    def is_busy( self, auto_apply=True ):
        if auto_apply:
            try:
                if self.analyzer.cancel_reports:
                    # this case is for testing only
                    dialog = self.analyzer.dialog
                    if dialog is None:
                        pass
                else:
                    dialog = self.analyzer.progress_dialog
                    if dialog is None:
                        pass
                    else:
                        text = dialog.button.cget('text')
                        if text == 'OK':
                            self.analyzer.progress_dialog.button.invoke()
            except:
                etb = ExceptionTracebacker()
                print(etb)
                pass
        return self.is_executing

    def report_callback_exception(self, exc, val, tb):
        # This method is to override the Tk method to be able
        # to report to the spawned console in windows application mode.

        if self.tmp_logger is None:
            self.tmp_logger = Logger()
        logger = self.tmp_logger

        etb = ExceptionTracebacker()
        msg = ( 'Overridden report_callback_exception: ' + str( etb )
                + '\n---- %s ----\n' % get_version_string(cpuid=True)
                + 'This message is saved either to "%s"\n' % logger.get_final_log_path()
                + 'or to "molass.log" in the analysis report folder.'  )

        logger.error( msg )
        self.tmp_logger = logger = Logger()     # this line invokes Logger's destructor
                                                # and recovers another logger in case
                                                # the execution continues
        """
            this is necessary for
            some modules such as scipy.interpolate seem to exit
            without giving any chance to the destructor
        """

        if not self.analyzer.logger_visible:
            try:
                MessageBox.showerror( "Tkinter Error", msg, parent=self.parent )
            except:
                # _tkinter.TclError: bad window path name ".!guimain"
                pass

    def adjust_geometry( self ):
        self.update()
        self.deiconify()
        if self.intial_geometry is None:
            self.geometry( adjusted_geometry( self.geometry() ) )
            self.intial_geometry = self.geometry()

    def build_window( self ):
        from molass_legacy.GuiParts.FileInfoTable import FileInfoTable    # must be imported after initialize_settings

        global radio_button_text_length

        lib_dir = os.path.dirname( os.path.dirname( os.path.abspath( __file__ ) ) )
        iconpath = lib_dir + '/_MOLASS/molamola.ico'
        self.iconbitmap(default=iconpath)           # this default argument key is for Windows only

        # TODO: recover the previous state
        # main_geometry = get_setting( 'main_geometry' )
        # print( main_geometry )
        # self.geometry( main_geometry )
        # self.geometry( default_main_geometry )

        # temp_label = DynamicLabel(self)
        # fixed_font = temp_label.get_fixed_font()
        # fixed_font = Font.Font(family="FixedSys", size=8 )
        # fixed_font = Font.Font( family="ＭＳ ゴシック", size=10 )
        # fixed_font = Font.Font( family="Lucida Console", size=10 )
        # fixed_font = Font.Font( family="DotumChe", size=10 )
        # fixed_font = Font.Font( family="Consolas", size=10 )
        self.fixed_font = Font.Font( family="Courier", size=9 )

        self.create_menus()

        # basic Layout
        top_space = Tk.Frame( self, height=20 )
        top_space.pack()

        upper_frame = Tk.Frame( self )
        upper_frame.pack( fill=Tk.BOTH, padx=20 )

        lower_frame = Tk.Frame( self )
        lower_frame.pack( fill=Tk.BOTH, expand=1, padx=20 )

        # layout params
        section_label_width = 20
        button_width = 12
        state_show_width = 23
        state_show_sizable_width    = 164
        state_show_sizable_height   = 20
        folder_entry_width          = 60

        grid_row = -1

        # Input Information --------------------------------------------------------
        grid_row += 1
        column_adjust  = Tk.Frame( upper_frame, width=160, height=0 )
        column_adjust.grid( row=grid_row, column=0 )

        input_label  = Tk.Label( upper_frame, text= 'Input', width=section_label_width,
                                    relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        input_label.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        in_folder_label  = Tk.Label( upper_frame, text= 'Xray Scattering Data Folder: ' )
        in_folder_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.in_folder = Tk.StringVar()
        self.in_folder.set( get_setting( 'in_folder' ) )
        self.in_folder_entry = FolderEntry( upper_frame, textvariable=self.in_folder, width=folder_entry_width,
                                            slimbutton=self.slimbutton,
                                            on_entry_cb=self.on_entry_in_folder )
        self.in_folder_entry.grid( row=grid_row, column=1, sticky=Tk.W )
        self.previous_in_folder = None
        self.previous_uv_file = None
        self.uv_file_info_changed = False
        self.cleared_uv_folder  = False

        file_extension_label  = Tk.Label( upper_frame, text= '  File Ext.: ' )
        file_extension_label.grid( row=grid_row, column=2, sticky=Tk.E )
        self.file_extension  = Tk.StringVar()
        self.file_extension.set( get_setting( 'file_extension' ) )
        file_extension_box = ttk.Combobox( upper_frame, textvariable=self.file_extension,
                                            width=5, justify=Tk.CENTER )
        file_extension_box[ 'values' ] = [ '*.dat', '*.csv', '*.int', '*.*' ]
        file_extension_box.grid( row=grid_row, column=3, sticky=Tk.W )

        self.file_extension.trace( "w", self.file_extension_tracer )

        grid_row += 1
        uv_folder_label  = Tk.Label( upper_frame, text= 'UV Absorbance Data Folder: ' )
        uv_folder_label.grid( row=grid_row, column=0, sticky=Tk.E )

        self.uv_folder = Tk.StringVar()
        self.uv_folder.set( get_setting( 'uv_folder' ) )
        self.uv_folder_entry = UvFolderEntry( upper_frame, textvariable=self.uv_folder, width=folder_entry_width,
                                            slimbutton=self.slimbutton,
                                            on_entry_cb=self.on_entry_uv_folder )
        self.uv_folder_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        uv_file_name_label  = Tk.Label( upper_frame, text= '  File Name: ' )
        uv_file_name_label.grid( row=grid_row, column=2, sticky=Tk.E )
        self.uv_file = Tk.StringVar()

        uv_file_frame = Tk.Frame( upper_frame )
        uv_file_frame.grid( row=grid_row, column=3, sticky=Tk.W )

        # TODO: refactoring with FileEntry
        self.uv_file_entry = Tk.Entry( uv_file_frame, textvariable=self.uv_file, width=30 )
        self.uv_file_entry.grid( row=0, column=0, sticky=Tk.W )
        if self.slimbutton:
            b3 = SlimButton( uv_file_frame, text='...', command=self.select_uv_file, width=18, height=22 )
        else:
            b3 = Tk.Button( uv_file_frame, text='...', command=self.select_uv_file )

        b3.grid( row=0, column=1, sticky=Tk.W )

        # DND bind
        self.uv_file_entry.register_drop_target("*")

        def dnd_handler( event ):
            self.select_uv_file(path=event.data)

        self.uv_file_entry.bind("<<Drop>>", dnd_handler)

        # disable UV data
        grid_row += 1
        self.disable_uv_data = Tk.IntVar()
        self.disable_uv_data.set(get_setting('disable_uv_data'))
        self.disable_uv_data_cb = Tk.Checkbutton( upper_frame, variable=self.disable_uv_data,
                                text='disable UV data', state=Tk.DISABLED )
        self.disable_uv_data_cb.grid(row=grid_row, column=2, columnspan=2, sticky=Tk.W, padx=5)
        self.disable_uv_data_tracing = True
        self.disable_uv_data_tracer_busy = False
        self.disable_uv_data.trace('w', self.disable_uv_data_tracer)

        # Conc. Factors
        from molass_legacy.GuiParts.ConcFactorsEntry import ConcFactorsEntry

        grid_row += 1
        self.cfs_entry = ConcFactorsEntry(upper_frame, grid_row)

        # pad
        grid_row += 2   # ConcFactorsEntry includes 2 rows
        """
        self.no_absorbance = Tk.IntVar()
        cb = Tk.Checkbutton( upper_frame, text='Use without Absorbance Data',
                    variable=self.no_absorbance )
        cb.grid( row=grid_row, column=1, sticky=Tk.W )
        """

        pad_frame = Tk.Label( upper_frame )
        pad_frame.grid( row=grid_row, column=0 )

        # Analysis Result Setting ----------------------------------------------
        grid_row += 1

        analysis_label  = Tk.Label( upper_frame, text= 'Output', width=section_label_width,
                                    relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        analysis_label.grid( row=grid_row, column=0, sticky=Tk.W )

        grid_row += 1
        an_folder_label  = Tk.Label( upper_frame, text= 'Analysis Result Folder: ' )
        an_folder_label.grid( row=grid_row, column=0, sticky=Tk.E )
        self.an_folder = Tk.StringVar()
        self.an_folder.set( get_setting( 'an_folder' ) )
        self.an_folder_frame = BlinkingFrame( upper_frame )
        self.an_folder_frame.grid( row=grid_row, column=1, sticky=Tk.W )
        self.an_folder_entry = FolderEntry( self.an_folder_frame, textvariable=self.an_folder, width=folder_entry_width,
                                            slimbutton=self.slimbutton,
                                            on_entry_cb=self.on_entry_an_folder )
        self.an_folder_entry.pack()
        self.an_folder_entry.bind('<FocusOut>', self.an_folder_focusout)
        self.an_folder_frame.objects = [self.an_folder_entry.entry]

        grid_row += 1
        analysis_name_label = Tk.Label( upper_frame, text='Subfolder: ' )
        analysis_name_label.grid( row=grid_row, column=0, sticky=Tk.E )
        analysis_frame = Tk.Frame( upper_frame )
        analysis_frame.grid( row=grid_row, column=1, sticky=Tk.W )
        self.analysis_name = Tk.StringVar()
        self.analysis_name.set( get_setting( 'analysis_name' ) )
        self.analysis_name_entry = Tk.Entry( analysis_frame, textvariable=self.analysis_name, width=20 )
        self.analysis_name_entry.grid( row=0, column=0, sticky=Tk.W )
        self.auto_number = Tk.IntVar()
        auto_number_ = get_setting( 'auto_number' )
        self.auto_number.set( auto_number_ )
        if False:
            self.auto_numbering_cb = Tk.Checkbutton( analysis_frame, variable=self.auto_number,
                                                text='Auto-number (un-check to reuse)' )
                                                # text='Auto-number (un-check this to reuse previous results)' )
            self.auto_numbering_cb.grid( row=0, column=1, sticky=Tk.W, padx=10 )

        if auto_number_ == 1 and not is_empty_val( self.an_folder.get() ):
            analysis_name, _ = self.make_analysis_folder( format_check_only=True )
            self.analysis_name.set( analysis_name )

        self.analysis_name_entry.bind( '<FocusOut>',  self.analysis_name_entry_focus_out )

        grid_row += 1
        result_book_label = Tk.Label( upper_frame, text='Book Name: ' )
        result_book_label.grid( row=grid_row, column=0, sticky=Tk.E )
        self.result_book = Tk.StringVar()
        result_book = get_setting( 'result_book' )
        self.result_book.set( result_book )
        self.result_book_entry = Tk.Entry( upper_frame, textvariable=self.result_book, width=20 )
        self.result_book_entry.grid( row=grid_row, column=1, sticky=Tk.W )

        grid_row += 1
        pad_frame = Tk.Label( upper_frame )
        pad_frame.grid( row=grid_row, column=0 )

        # File Information Table -----------------------------------------------
        grid_row += 1
        grid_row_for_analysis_button = grid_row
        table_label  = Tk.Label( upper_frame, text= 'Xray Scattering Data Files', width=section_label_width,
                                    relief=Tk.FLAT, fg=label_fg_color, bg=label_bg_color )
        table_label.grid( row=grid_row, column=0, sticky=Tk.W )

        fig_table_frame = Tk.Frame( lower_frame )
        fig_table_frame.pack(fill=Tk.X, expand=1, pady=10)

        # navigation button frame
        nav_frame = Tk.Frame(fig_table_frame)
        nav_frame.pack( side=Tk.BOTTOM, fill=Tk.X, expand=1, pady=10)

        self.message_frame = Tk.Frame( fig_table_frame )
        self.message_frame.pack( side=Tk.BOTTOM, fill=Tk.X, expand=1)   # Tk.BOTTOM required!

        table_height = 150 if self.high_dpi else 100
        self.file_info_table = FileInfoTable( fig_table_frame, self, self.loader, self.fixed_font, height=table_height )
        self.file_info_table.pack( side=Tk.RIGHT, fill=Tk.BOTH, expand=1, anchor=Tk.N)

        figsize = (6,8) if self.high_dpi else (4,5)
        self.fig_frame = OutlineFigure(fig_table_frame, self, file_info_table=self.file_info_table, figsize=figsize)
        self.fig_frame.pack(side=Tk.LEFT, fill=Tk.BOTH)

        # fill=Tk.BOTH would make the Scrollbar too long in some cases

        self.btn_frame = Tk.Frame(upper_frame)
        self.btn_frame.grid(row=grid_row, column=1, columnspan=3, sticky=Tk.E)

        self.refresh_button = Tk.Button( self.btn_frame, text='Refresh', command=self.refresh )
        self.refresh_button.grid(row=0, column=0)
        self.presync_button = Tk.Button( self.btn_frame, text='3D View', command=self.show_threedim_dialog )
        self.presync_button.grid(row=0, column=1, padx=10)
        self.restrict_button = Tk.Button( self.btn_frame, text='Data Range', command=self.show_datarange_dialog )
        self.restrict_button.grid(row=0, column=2)
        self.rankview_menu = RankViewMenu(self.btn_frame, self)
        self.rankview_menu.grid(row=0, column=3, padx=10)
        self.baseline_button = Tk.Button( self.btn_frame, text='Baseline', command=self.show_lpm_inspector )
        self.baseline_button.grid(row=0, column=4)
        if is_developing_version():
            self.debug_button = Tk.Button( self.btn_frame, text='Dev Trial', command=self.show_dev_trial_entry )
            self.debug_button.grid(row=0, column=5, padx=10)

        # Analysis start button ---------------------------------------------------
        nav_btn_frame = Tk.Frame(nav_frame)
        nav_btn_frame.pack(side=Tk.RIGHT)

        if False:
            preproc_button = Tk.Button(nav_frame, text="◀ Preprocess", command=self.back_to_preproc)
            preproc_button.pack(side=Tk.LEFT)

        exit_button = Tk.Button(nav_frame, text="▼ Exit", command=self.quit)
        exit_button.pack(side=Tk.LEFT, padx=10)

        button_frame = Tk.Frame( nav_btn_frame )
        button_frame.pack()

        self.v2_menu = V2Menu(button_frame, self)
        self.v2_menu.grid(row=0, column=0, padx=5)

        self.analysis_button = Tk.Button( button_frame, text="▶ Analysis start", command=self.analyze_all )
        self.analysis_button.grid( row=0, column=1, padx=5 )
        self.fully_automatic  = Tk.IntVar()
        self.fully_automatic.set(get_setting('fully_automatic'))
        self.fully_automatic_cb = Tk.Checkbutton( button_frame, text="fully automatic", variable=self.fully_automatic )
        self.fully_automatic_cb.grid( row=0, column=2, padx=5 )
        self.buttons = [self.refresh_button,
                            self.restrict_button, self.presync_button, self.rankview_menu, self.baseline_button,
                            self.v2_menu,
                            self.analysis_button, self.fully_automatic_cb,
                            self.disable_uv_data_cb, self.cfs_entry,
                            exit_button]

        self.enable_synthesized_lrf = get_setting('enable_synthesized_lrf')
        if self.enable_synthesized_lrf:
            self.synthesized_lrf  = Tk.IntVar()
            self.synthesized_lrf.set(get_setting('synthesized_lrf'))
            self.synthesized_lrf_cb = Tk.Checkbutton( button_frame, text="synthesized LRF", variable=self.synthesized_lrf )
            self.synthesized_lrf_cb.grid( row=0, column=3, padx=5 )
            self.buttons.append(self.synthesized_lrf_cb)

        self.update_button_states()
        self.update_analysis_button_state()

        # Plot results button ---------------------------------------------------
        self.plot_button = Tk.Button( button_frame, text="Plot Results", command=self.plot_results )
        self.plot_button.grid( row=0, column=1, padx=5 )
        self.update_plot_button_state()

        self.in_folder_has_been_changed = False
        self.settings_have_been_changed = False

        # it seemd that '<Control-c>' should be bound from the top-level.
        self.bind( '<Control-c>', self._on_control_c )
        self.bind( '<Button-1>', self._on_click )
        # self.bind( '<Return>', self._on_enter )
        # self.bind( '<Control-a>', self._on_enter )
        self.bind( '<Escape>', self._on_escape )
        self.in_folder_entry.bind( '<FocusIn>', self.on_in_folder_entry_focus_in )
        self.in_folder_entry.bind( '<FocusOut>',  self.on_in_folder_entry_focus_out )

        # message frame
        self.guide_message = Tk.Label(self.message_frame, bg='white')
        self.guide_message.pack(fill=Tk.X, pady=10)

        self.set_state_guide_message("Specify input/output folders")

        # getting ready to show
        self.adjust_geometry()

        invoke_on_entry = False
        if not is_empty_val( self.in_folder.get() ):
            invoke_on_entry = self.ask_resume()

        # file_info_table does not grow in size after geometry setting
        self.file_info_table.refresh()

        if invoke_on_entry:
            self.on_entry_in_folder()

        uv_folder_ = self.uv_folder.get()
        if not is_empty_val( uv_folder_ ):
            if is_empty_val( self.uv_file.get() ):
                uv_file = get_uv_filename( uv_folder_ )
                self.uv_file.set( uv_file )

        self.datarange_dialog = None
        self.protocol( "WM_DELETE_WINDOW", self.quit )

    def ask_resume(self):
        self.update()
        ret = MessageBox.askyesno("Reload Confirmation",
            'Previous input was from\n'
            '%s\n'
            'Would you like to have it loaded?\n'
            % self.in_folder.get(), parent=self,
            )
        if ret:
            invoke_on_entry = True
        else:
            self.in_folder.set(None)
            set_setting("in_folder", None)
            invoke_on_entry = False
        return invoke_on_entry

    def set_state_guide_message(self, guide_message, fg='black'):
        self.message_frame.update()
        try:
            self.guide_message.config(text=guide_message.replace('\n', ' '), fg=fg)
        except:
            # _tkinter.TclError: invalid command name ".!guimain.!frame3.!frame.!frame2.!label"
            log_exception(self.tmp_logger, "self.guide_message.config(...): ")

    def file_extension_tracer( self, *args ):
        set_setting( 'file_extension', self.file_extension.get() )
        self.on_entry_in_folder()

    def refresh( self, keeping_setting_items=None):
        print( 'refresh' )
        self.keeping_setting_items = keeping_setting_items
        self.previous_in_folder = None  # this will induce clear_temporary_settings()
        self.in_folder_entry.on_entry(refresh=True)
        # self.uv_folder_entry.on_entry()

        if self.pre_recog.restrict_info_changed():
            yn = MessageBox.askyesno("Question",
                            "Data Range info seems to have been changed mannually.\n"
                            "Do you wish reset it to the default?",
                            parent=self)
            if yn:
                self.pre_recog.reset_restrict_info()

    def show_datarange_dialog( self ):
        from molass_legacy.Trimming import DataRangeDialog
        range_editor_info = get_setting('range_editor_info')
        decomp_editor_info = get_setting('decomp_editor_info')

        if range_editor_info is not None or decomp_editor_info is not None:
            ok = MessageBox.askyesno( 
                            'Confirmation Question',
                            'Previously memorized analysis range info exists.\n'
                            'Applying a new data restriction will reset that info to default.\n'
                            'Are you sure to proceed?',
                            parent=self
                            )
            if not ok:
                return

        self.datarange_dialog = DataRangeDialog(self.parent, self.pre_recog)
        self.datarange_dialog.show()

        self.draw_figure()

    def has_datarange_dialog(self):
        return self.datarange_dialog is not None

    def create_menus( self ):
        import molass_legacy._MOLASS.SerialSettings
        menubar = Tk.Menu( self )
        self.config( menu=menubar )
        self.menubar = menubar

        menu0 = Tk.Menu( menubar, tearoff=0 )
        self.recent_menu = Tk.Menu( menubar, tearoff=0 )

        menubar.add_cascade( label="Folder", menu=menu0 )
        menu0.add_cascade( label="Recently used data folders", menu=self.recent_menu )
        self.recent_folders = RecentFolders( size=20, settings=molass_legacy._MOLASS.SerialSettings)
        self.recent_menu_add()
        menu0.add_command( label="Exit", command=self.quit )

        menu1 = GuiSettingsMenu( self, menubar )
        self.menu3 = GuiSecToolsMenu( self, menubar )
        self.menuf = GuiMctToolsMenu( self.parent, self, menubar )
        self.menu4 = GuiDenssToolsMenu( self, menubar )
        self.menu5 = GuiTutorialsMenu( self, menubar )
        menu6 = GuiReferencesMenu( self, menubar )
        menu7 = GuiDevelopmentMenu( self, menubar )

    def recent_menu_add( self ):
        for folder in self.recent_folders.get_sorted_list():
            path = folder[1]
            # print( 'add recent menu', path )
            self.recent_menu.add_command( label=path, command=lambda path_=path: self.set_in_folder( path_ ) )

        self.update()

    def recent_menu_delete( self ):
        for folder in self.recent_folders.get_sorted_list():
            path = folder[1]
            # print( 'deleting menu ', path )
            try:
                self.recent_menu.delete( path )
            except:
                print( 'error deleting menu ', path )

    def askdirectory( self, entry_variable ):
        entered_path = entry_variable.get()
        # print 'entered_path=', entered_path
        dir_ = os.path.dirname( entered_path ).replace( '/', '\\' )
        # print 'dir_=', dir_
        f = FileDialog.askdirectory( initialdir=dir_, parent=self )
        return f

    def set_in_folder( self, path ):
        # print( 'set_in_folder', path )
        self.in_folder.set( path )
        self.on_entry_in_folder()
        self.settings_have_been_changed = True

    def set_uv_folder( self, path, filename ):
        # print( 'set_uv_folder', path )
        self.uv_folder.set( path )
        self.on_entry_uv_folder( filename=filename )
        self.settings_have_been_changed = True

    def select_uv_file(self, path=None):
        if path is None:
            f = FileDialog.askopenfilename( initialdir=self.uv_folder.get(), parent=self )
            if not f:
                return
        else:
            f = path

        dir_, file = os.path.split( f )
        self.on_entry_uv_file( dir_, file  )
        self.settings_have_been_changed = True

    def update_button_states(self):
        state = Tk.NORMAL if (self.dataset_is_ready and not self.is_executing) else Tk.DISABLED
        for k, btn in enumerate(self.buttons):
            btn.config(state=state)

        disable_uv_data = get_setting('disable_uv_data')
        if not disable_uv_data:
            if is_empty_val(self.uv_file.get()):
                for w in [self.cfs_entry, self.disable_uv_data_cb]:
                    w.config(state=Tk.DISABLED)

        self.update_menu_states()
        self.update()

    def update_menu_states(self):
        self.menu3.update_states()
        self.menuf.update_states()
        self.menu4.update_states()
        self.menu5.update_states()

    def update_analysis_button_state( self ):
        if self.is_executing:
            state_ = Tk.DISABLED
        else:
            if self.dataset_is_ready:
                num_data_rows = self.file_info_table.table.cells_array.shape[0] - 1
                use_xray_conc = get_setting( 'use_xray_conc' ) == 1
                use_mtd_conc = get_setting( 'use_mtd_conc' ) == 1
                if use_xray_conc or use_mtd_conc:
                    state_ = Tk.NORMAL if num_data_rows > 0 else Tk.DISABLED
                else:
                    state_ = Tk.NORMAL if num_data_rows > 0 and self.uv_file_exists else Tk.DISABLED
            else:
                state_ = Tk.DISABLED
        try:
            self.analysis_button.config( state=state_ )
            self.update()
        except:
            # _tkinter.TclError: invalid command name ".!guimain.!frame3.!frame.!frame.!frame.!frame.!button"
            pass

    def update_all_button_states(self):
        self.update()
        self.update_button_states()
        self.update_analysis_button_state()
        self.update_plot_button_state()
        self.update()

    def on_entry_in_folder( self, refresh=False ):
        # refresh is currently not used

        self.uv_file_info_changed = False
        self.dataset_is_ready = False
        self.config( cursor='wait' )
        self.update_all_button_states()
        self.uv_file_exists = False

        self.conc_serie = None
        in_folder_ = self.in_folder.get()

        if in_folder_ != self.previous_in_folder:
            kept_items = None
            if refresh and self.keeping_setting_items is not None:
                kept_items = []
                for item in self.keeping_setting_items:
                    kept_items.append(get_setting(item))
            self.on_in_folder_change()
            self.load_start_time = time.time()
            if kept_items is not None:
                for item, value in zip(self.keeping_setting_items, kept_items):
                    set_setting(item, value)
            self.serial_data = None
            self.measured_data = None
            self.previous_in_folder = in_folder_
            self.cleared_uv_folder  = False

        self.recent_menu_delete()
        self.recent_folders.add( in_folder_ )
        self.recent_menu_add()

        set_setting( 'in_folder', in_folder_ )
        self.file_info_table.refresh()
        self.update()

        self.config( cursor='' )

        if self.file_info_table.num_rows == 0:
            self.set_in_folder_error()
            self.update()   # necessary for the file extension combobox to shrink
            MessageBox.showerror( "Xray Scattering Data Folder Error", "No Data File in '%s'" % in_folder_, parent=self )
            self.set_in_folder_error()  # call it again to focus
            return

        # need to restore fg if the case is after the error above
        self.in_folder_entry.config( fg='black' )

        self.loader.reset_input_status()
        if self.tmp_logger is None:
            self.tmp_logger = Logger()  # this logger logs into string stream
        else:
            # self.tmp_logger should have been created after do_analysis
            pass

        if self.parent_pid is not None:
            from molass_legacy.KekLib.CleanerThread import CleanerThread
            with_parent_pid = " with parent_pid=%d" % self.parent_pid
            self.cleaner = CleanerThread(self.parent_pid)
        else:
            with_parent_pid = ""
            self.cleaner = None
        self.tmp_logger.info('logging started on_entry_in_folder in %s%s', get_version_string(cpuid=True), with_parent_pid)

        if self.uv_file_info_changed:
            # keep the current uv file info
            self.loader.reset_current_status()
            self.after(500, self.prepare_serial_data)
        else:
            filename = get_uv_filename( in_folder_ )
            if filename is None:
                filename = get_mtd_filename( in_folder_ )
                if filename is None:
                    if not self.cleared_uv_folder:
                        # don't do this for the same in_folder_
                        self.uv_folder.set( '' )
                        self.uv_file.set( '' )
                        set_setting( 'uv_folder',   None )  # clear
                        set_setting( 'uv_file',     None )  # clear
                        self.cleared_uv_folder = True
                        set_setting( 'use_xray_conc', 1 )
                        from molass_legacy._MOLASS.SerialSettings import do_xray_conc_temporary_settings
                        do_xray_conc_temporary_settings()
                        self.after(500, self.prepare_serial_data)
                else:
                    # 
                    folder, file = os.path.split(filename)
                    self.uv_folder.set( folder )
                    self.uv_file.set( file )

                    set_setting( 'use_xray_conc', 0 )
                    set_setting( 'use_mtd_conc', 1 )
                    self.use_mtd_conc = True
                    set_setting( 'mtd_file_path', filename )
                    # self.fully_automatic.set(0)
                    from molass_legacy._MOLASS.SerialSettings import do_microfluidic_temporary_settings
                    do_microfluidic_temporary_settings()
                    self.after(500, self.prepare_serial_data)
            else:
                self.set_uv_folder( in_folder_, filename )
                set_setting( 'use_xray_conc', 0 )
                # verify the need to call self.prepare_serial_data
                # self.after(500, self.prepare_serial_data)

        self.lacking_q_values_warned = False
        self.fig_frame.clear_figures()
        self.set_state_guide_message("Waiting for the completion\nof data loading.")
        self.update()
        # update states
        self.update_uv_folder_entry_state()
        self.settings_have_been_changed = True

    def clear_entries(self):
        self.cfs_entry.reset_entries()
        clear_temporary_settings()
        self.update()

    def on_in_folder_change(self):
        self.clear_entries()

        if self.disable_uv_data_tracer_busy:
            """
            do nothing when called during the tracer process.
            """
            pass
        else:
            """
            this is the case when the input folder has been changed
            after disabling the previous folder.
            """
            self.disable_uv_data_tracing = False
            self.disable_uv_data.set(get_setting('disable_uv_data'))
            self.update()
            self.disable_uv_data_tracing = True

    def set_in_folder_error( self ):
        self.in_folder_entry.config( fg='red' )
        self.in_folder_entry.focus_force()

    def on_entry_uv_folder( self, filename=None ):
        if self.loader.is_busy:
            # TODO : better to get rid of this case
            if self.tmp_logger is not None:
                self.tmp_logger.info('on_entry_uv_folder is skipped because the loader is busy.')
                return

        uv_folder_ = self.uv_folder.get()

        if filename is None:
            filename = get_uv_filename( uv_folder_ )
            if filename is None:
                self.set_uv_folder_error()
                MessageBox.showerror( "Absorbance Folder Error", "No absorbance file in '%s'." % uv_folder_, parent=self )
                self.uv_file.set( "<Not found>" )
                return

        self.uv_file_info_changed = True
        self.uv_file.set( filename )
        set_setting( 'uv_folder', uv_folder_ )
        set_setting( 'uv_file', filename )
        self.uv_file_exists = True

        data_folder = get_setting( 'in_folder' ).replace( '\\', '/' )
        conc_folder = get_setting( 'uv_folder' ).replace( '\\', '/' )
        conc_file   = get_setting( 'uv_file'   )
        if conc_file.find( '*' ) >= 0:
            conc_file = None

        # self.analysis_button.config( state=Tk.NORMAL )

        if not is_empty_val(self.previous_uv_file):
            self.set_state_guide_message("Re-loading the data")

        self.previous_uv_file = filename
        self.update_analysis_button_state()
        self.loader.load_from_folders( data_folder, uv_folder=conc_folder, uv_file=conc_file )
        # serial_data construction will be delayed until self.analyze_all
        # to keep this thread running without waiting here

        self.settings_have_been_changed = True
        self.after(500, self.prepare_serial_data)

    def set_uv_folder_error( self ):
        self.uv_folder_entry.config( fg='red' )

    def on_entry_uv_file( self, dir_, file ):
        self.uv_file.set( file )
        self.uv_folder.set( dir_ )

        if not self.uv_folder_entry.check():
            return

        if not self.uv_file_check():
            return

        self.set_uv_folder( dir_, file )

    def uv_file_check( self ):
        # TODO: include these checks in a custom widget class
        ok_ = False
        uv_file = self.uv_file.get()
        uv_folder = self.uv_folder.get()
        if uv_file.find( '*' ) >= 0:
            file_ = get_uv_filename(uv_folder, glob=uv_file)
            if file_ is None:
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "No Absorbance File in '%s'" % uv_folder, parent=self )
            else:
                self.uv_file_entry.config( fg='black' )
                ok_ = True
        else:
            uv_filepath = os.path.join( uv_folder, uv_file );
            if is_empty_val( uv_file ):
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "The Absorbance File is required.", parent=self )
            elif not os.path.exists( uv_filepath ):
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "'%s' does not exist." % uv_file, parent=self )
            elif not os.path.isfile( uv_filepath ):
                self.set_uv_file_error()
                MessageBox.showerror( "Absorbance File Error", "'%s' is not a file." % uv_file, parent=self )
            else:
                self.uv_file_entry.config( fg='black' )
                ok_ = True

        if not ok_:
            self.uv_file_entry.focus_force()

        return ok_

    def set_uv_file_error( self ):
        self.uv_file_entry.config( fg='red' )

    def on_in_folder_entry_focus_in( self, event ):
        self.in_folder_ext_state = self.in_folder.get() + self.file_extension.get()

    def on_in_folder_entry_focus_out( self, event ):
        # print( 'on_in_folder_entry_focus_out' )
        in_folder = self.in_folder.get()
        file_extension = self.file_extension.get()
        in_folder_ext_state = in_folder + file_extension
        if in_folder_ext_state == self.in_folder_ext_state:
            return

        uv_folder = self.uv_folder.get()
        if is_empty_val( in_folder ) or not is_empty_val( uv_folder ):
            return

        self.on_entry_in_folder()

    def on_entry_an_folder( self ):
        # print( 'on_entry_an_folder' )
        self.an_folder_frame.stop()
        self.update()

        if not self.an_folder_check():
            return

        # self.an_folder_entry.config(fg=None)
        self.config( cursor='wait' )    # doesn't seem to work
        self.update()
        an_folder_ = self.an_folder.get()
        set_setting( 'an_folder', an_folder_ )
        self.settings_have_been_changed = True
        self.update_plot_button_state()
        if self.serial_data is not None:
            self.draw_figure()
        self.config( cursor='' )
        self.make_it_more_clever()

    def an_folder_focusout( self, *args ):
        if not is_empty_val(self.an_folder.get()):
            if self.an_folder_frame.is_blinking():
                self.on_entry_an_folder()

    def an_folder_check( self ):
        # TODO: include these checks in a custom widget class
        ok_ = False
        an_folder = self.an_folder.get()
        if is_empty_val( an_folder ):
            self.set_an_folder_error()
            MessageBox.showerror( "Analysis Folder Error", "The Analysis Result Folder is required.", parent=self )
        else:
            from molass_legacy.GuiParts.NonAsciiPaths import nonascii_path_check
            if not nonascii_path_check(an_folder, self):
                return ok_

            self.an_folder_frame.stop()
            if not os.path.exists( an_folder ):
                self.set_an_folder_error()
                yn = MessageBox.askyesno( "Analysis Folder Error",
                            "'%s' does not exist.\nWould you like to make the folder?" % an_folder, icon="warning", parent=self )
                if yn:
                    try:
                        mkdirs_with_retry( an_folder )
                    except:
                        pass
                    if os.path.exists( an_folder ):
                        ok_ = True
            elif not os.path.isdir( an_folder ):
                self.set_an_folder_error()
                MessageBox.showerror( "Analysis Folder Error", "'%s' is not a folder." % an_folder, parent=self )
            elif self.check_if_like_report_folder( an_folder ):
                self.set_an_folder_error()
                yn = MessageBox.askyesno( "Analysis Folder Error",
                                            ( "'%s' seems like a used sub-folder.\n"
                                             + "Are you sure this is an appropriate selection?" )
                                            % an_folder, parent=self )
                if yn:
                    ok_ = True
            else:
                ok_ = True

        if ok_:
            self.an_folder_entry.config( fg='black' )
        else:
            self.an_folder_entry.focus_force()

        return ok_

    def check_if_like_report_folder( self, an_folder ):
        logfile = an_folder + '/molass.log'
        repfile = an_folder + '/analysis_report.xlsx'
        return os.path.exists( logfile ) and  os.path.exists( repfile )

    def set_an_folder_error( self ):
        self.an_folder_entry.config( fg='red' )

    def new_analysis( self ):
        reuse = get_setting("reuse_analysis_folder")
        if reuse:
            analysis_name = get_setting('analysis_name')
            analysis_folder = get_setting('analysis_folder')
            if analysis_folder is None:
                pass
            else:
                return analysis_name, analysis_folder

        try:
            analysis_name, analysis_folder = self.make_analysis_folder()
            self.analysis_name.set( analysis_name )
            set_setting( 'analysis_name', analysis_name )
            set_setting( 'analysis_folder', analysis_folder )
            return analysis_name, analysis_folder
        except:
            print( sys.exc_info() )
            return

    def get_analysis_name(self):
        # this method is for use in testing
        an_folder = self.an_folder.get()
        analysis_name = self.analysis_name.get()
        return self.auto_number_analysis_name(an_folder, analysis_name)

    def make_analysis_folder( self, format_check_only=False ):
        an_folder = self.an_folder.get()
        analysis_name = self.analysis_name.get()
        self.auto_number_analysis_name( an_folder, analysis_name ) # this is just for check if it is propperly named

        def get_analysis_folder( out_folder, analysis_name ):
            if is_empty_val( out_folder ):
                return None
            else:
                analysis_folder = os.path.join( out_folder, analysis_name )
                return analysis_folder.replace( '\\', '/' )

        analysis_folder = get_analysis_folder( an_folder, analysis_name )

        if not format_check_only:

            if self.auto_number.get() == 1:
                if not is_empty_val( analysis_folder ):
                    while os.path.exists( analysis_folder ) and not is_almost_empty_dir( analysis_folder ):
                        analysis_name = self.auto_number_analysis_name( an_folder, analysis_name )
                        analysis_folder = get_analysis_folder( an_folder, analysis_name )

            # print( 'make_analysis_folder: analysis_folder=', analysis_folder, type(analysis_folder) )
            if not is_empty_val( analysis_folder ) and not os.path.exists( analysis_folder ):
                mkdirs_with_retry( analysis_folder )

        return analysis_name, analysis_folder

    def auto_number_analysis_name( self, an_folder, analysis_name ):
        m = analysis_name_re.match( analysis_name )
        if m:
            first   = m.group( 1 )
            number  = m.group( 2 )
            third   = m.group( 3 )
            len_ = len( number )
            name_ = first + '%0*d' % ( len_, int( number ) ) + third
            if not is_empty_val( an_folder ):
                folder = an_folder + '/' + name_
                if os.path.exists( folder ) and not is_almost_empty_dir( folder ):
                    name_ = first + '%0*d' % ( len_, int( number ) + 1 ) + third
            self.analysis_name_entry.config( fg='black' )
            # print( 'new name=', name_ )
            return name_
        else:
            self.analysis_name_entry.config( fg='red' )
            MessageBox.showerror(
                'Naming Error',
                "Analysis Name '%s' is not expected. Name it with a numeric postfix such as 'analysis-000'." % analysis_name,
                parent=self,
                )
            assert( False )

    def analysis_name_entry_focus_out( self, event ):
        # print( 'analysis_name_entry_focus_out' )
        self.update_plot_button_state()

    def update_analysis_folder(self):
        analysis_folder = get_setting('analysis_folder')
        _, analysis_folder_ = self.make_analysis_folder(format_check_only=True)
        if analysis_folder != analysis_folder_:
            self.new_analysis()

    def prepare_serial_data( self ):
        self.tmp_logger.info('preparing serial data.')
        self.prepare_start_time = time.time()
        self.serial_data = self.loader.get_current_object()
        # note that data are not yet deepcopied at this stage
        self.abnomality_dialog = None
        if not self.use_mtd_conc:
            exclude_abnormality( self.serial_data, self.file_info_table, self.tmp_logger, dialog=self )

        if self.serial_data.has_excluded_xray_elutions:
            self.loader.memorize_exclusion( self.serial_data )

        if self.serial_data.is_serial():
            self.pre_recog = PreliminaryRecognition(self.serial_data)
        else:
            self.pre_recog = None
            return

        try:
            self.fig_frame.set_data(self.pre_recog)
        except Exception as exc:
            if not self.use_mtd_conc:
                raise exc

        eno = self.serial_data.xray_curve.primary_peak_i

        try:
            self.file_info_table.select_row(eno)
        except:
            log_exception(self.tmp_logger, "prepare_serial_data: self.file_info_table.select_row(eno)")

        self.draw_figure(eno)
        self.cfs_entry.update_path_length()

        self.dataset_is_ready = True
        preparing_time = time.time() - self.prepare_start_time
        self.tmp_logger.info('data set is ready. it took %.3g seconds for preparing.', preparing_time)
        self.update_button_states()

        self.measured_data = MeasuredData(None, sd=self.serial_data, pre_recog=self.pre_recog)
        updated = self.measured_data.update_picking_params(self.tmp_logger)
        # TODO: update self.fig_frame
        if updated:
            self.fig_frame.update_elution_curve()
        self.tmp_logger.info('created measured_data.')

        if not self.lacking_q_values_warned:
            if get_setting('found_lacking_q_values'):
                self.lacking_q_values_warned = True
                MessageBox.showwarning( "Lacking Q-values",
                    "There have been found lacking Q-values\n"
                    "in some of the input files.\n"
                    "Make sure to confirm them in the molass.log",
                    parent=self )

    def re_prepare_serial_data( self ):
        save_items = ["uv_device_no", "path_length", "beamline_name", "xray_baseline_type"]   # any other?
        saved_values = []
        for name in save_items:
            saved_values.append(get_setting(name))      # save

        self.clear_entries()

        for name, value in zip(save_items, saved_values):
            set_setting(name, value)                    # restore

        if self.dataset_is_ready:
            self.dataset_is_ready = False
            self.update_all_button_states()
            self.tmp_logger.info("re-preparing serial data after user's settings change.")
            self.after(500, self.prepare_serial_data)

    def is_reloadable(self):
        return self.dataset_is_ready

    def draw_figure( self, selected=None ):
        if self.serial_data is None or not self.serial_data.is_serial():
            return

        self.update()   # seems to be required to be sure to get the current self.an_folder value

        if is_empty_val( self.an_folder.get() ):
            self.set_state_guide_message("Specify output folder.")
            self.an_folder_frame.start()
            # self.an_folder_entry.config(bg='yellow')
            return

        self.fig_frame.draw_figure(selected)
        self.set_state_guide_message('Press [▶ Analysis start] or click to select.', fg='green')
        self.an_folder_frame.stop()
        self.update()   # seems to be required sometimes (not always)

    def wait_until_the_data_is_ready( self ):
        while self.loader.is_busy:
            self.update()
            print('waiting for the completion of data loading')
            time.sleep(1)

    def detected_abnomality( self ):
        if self.abnomality_dialog is None:
            ret = False
        else:
            self.tmp_logger.warning( 'detected abnomality(bubles?) in the input data.' )
            ret = True
        return ret

    def get_range_type( self ):
        return get_setting( 'range_type' )

    def analyze_all( self ):
        from molass_legacy.SerialAnalyzer.AbnormalityCheck import update_abnormality_fix_state

        if not self.check_all_setting_info():
            self.analysis_button.grid( row=0, column=0, padx=5 )
            return

        try:
            analysis_name, analysis_folder = self.make_analysis_folder()
            if os.path.exists( analysis_folder ) and not is_almost_empty_dir( analysis_folder ):
                self.new_analysis()
        except:
            print( sys.exc_info() )
            return

        fully_automatic = self.fully_automatic.get()
        if fully_automatic:
            ret = MessageBox.askyesno("Fully Automatic Run Confirmation",
                    '"Fully automatic" execution may take several minutes.\n'
                    + 'Are you sure to proceed?',
                    parent=self)
            if not ret:
                return

        # TODO: if using the same in_folder, ask if the previous params should be retained or discarded

        self.analysis_name.set( analysis_name )

        set_setting( 'analysis_folder', analysis_folder )
        set_setting( 'analysis_name', analysis_name )
        set_setting( 'result_book', self.result_book.get() )
        set_setting( 'fully_automatic', fully_automatic )
        if self.enable_synthesized_lrf:
            set_setting( 'synthesized_lrf', self.synthesized_lrf.get() )

        self.update_setting_info()
        self.update_plot_button_state()

        self.analyzer.change_log_to(analysis_folder)

        update_abnormality_fix_state(self.pre_recog.get_pre_recog_copy(), self.file_info_table, self.analyzer.app_logger, dialog=self)

        def v1_exec_closure():
            self.set_state_guide_message("Preparing for Mapping Dialog.")
            self.analyzer.do_analysis(self.serial_data, self.pre_recog, self.measured_data, analysis_folder, analysis_name)
            self.set_state_guide_message("V1 manipulation done.")

        self.exec_wrapper(v1_exec_closure, "do_analysis")

    def exec_wrapper(self, exec_closure, exec_name):
        # this wrapper is intended to unify the button state control

        self.is_executing = True
        self.update_button_states()
        self.update_analysis_button_state()
        self.update()

        exec_closure()

        self.reset_tmp_logger(exec_name)
        self.is_executing = False
        try:
            self.update_button_states()
            self.update_analysis_button_state()
        except:
            """
            _tkinter.TclError: invalid command name ".!guimain.!frame3.!frame.!frame.!frame.!frame.!checkbutton"
            probabply harmless
            """
            pass

    def reset_tmp_logger(self, process_name):
        self.tmp_logger = Logger()
        self.tmp_logger.info("tmp_logger has been created after %s in " % process_name + get_version_string(cpuid=True))

    def check_all_setting_info( self ):
        # for method in [ self.in_folder_entry.check, self.uv_folder_entry.check, self.an_folder_check, self.check_atsas_executability ]:
        for method in [ self.in_folder_entry.check, self.uv_folder_entry.check, self.an_folder_check ]:
            ok_ = method()
            if not ok_:
                return False

        return True

    def check_atsas_executability(self):
        from molass_legacy.ATSAS.ExecCheck import atsas_exec_check
        return atsas_exec_check(self)

    def update_setting_info( self, clear_temp=False ):
        set_setting( 'in_folder', self.in_folder.get() )
        set_setting( 'uv_folder', self.uv_folder.get() )
        set_setting( 'uv_file',   self.uv_file.get() )
        set_setting( 'an_folder', self.an_folder.get() )
        set_setting( 'auto_number', self.auto_number.get() )
        self.cfs_entry.apply_entries()
        self.save_settings(clear_temp=clear_temp)

    def update_uv_folder_entry_state( self ):
        self.use_xray_conc = get_setting( 'use_xray_conc' ) == 1
        self.use_mtd_conc = get_setting( 'use_mtd_conc' ) == 1

        uv_folder_entry = self.uv_folder_entry.entry
        if self.use_xray_conc or self.use_mtd_conc:
            in_folder = self.in_folder.get()
            if not is_empty_val( in_folder ):
                set_setting( 'file_extension', self.file_extension.get() )
                # TODO: make this symmetric with load_from_folders
                self.loader.load_xray_data_only( in_folder )
                self.loader.wait_until_ready()
            if self.use_xray_conc:
                if self.loader.has_enough_num_files():
                    text = 'No UV data; using Xray-proportional data instead'
                else:
                    # self.fig_frame.clear_figures("Guinier analysis only is\navaiblable for non-serial data.")
                    text = 'No UV data; using non-serial data'
                uv_folder_entry.config( state=Tk.NORMAL )
                uv_folder_entry.insert(0, text)
                uv_folder_entry.config( state='readonly' )
                uv_folder_entry.config( fg='orange' )
                uv_folder_entry.config( justify='center' )
                self.uv_file_entry.config( state=Tk.DISABLED )
                normal_state = False
            else:
                normal_state = True
        else:
            normal_state = True

        if normal_state:
            uv_folder_entry.config( state=Tk.NORMAL )
            uv_folder_entry.config( fg='black' )
            uv_folder_entry.config( justify='left' )
            self.uv_file_entry.config( state=Tk.NORMAL )

        self.update_analysis_button_state()

    def doing_mfc(self):
        return self.use_mtd_conc == 1

    def plot_results( self ):
        print( 'plot_results' )
        from molass_legacy.GuinierAnalyzer.SimpleGuinierAnalyzer import SimpleGuinierAnalyzer
        from molass_legacy.GuinierAnalyzer.SimpleGuinierFolderAnalyzer import SimpleGuinierFolderAnalyzer
        in_folder = self.in_folder.get()
        sg_analyzer = SimpleGuinierAnalyzer( self.maintenance_log )
        report_folder = '/'.join( self.maintenance_log.split('/')[:-2]  )
        fo_analyzer = SimpleGuinierFolderAnalyzer( self, -1, in_folder,
                        sg_analyzer.folder_infos[0],
                        sg_analyzer.x, sg_analyzer.z, sg_analyzer.r,
                        sg_analyzer.type_names,
                        report_folder=report_folder )

        fo_analyzer.show()

    def update_plot_button_state( self ):
        maintenance_mode = get_setting( 'maintenance_mode' )
        if maintenance_mode == 0:
            self.plot_button.grid_forget()
            self.maintenance_log = None
        else:
            self.plot_button.grid( row=0, column=1, padx=5 )
            self.maintenance_log = self.make_maintenance_log_path()
        if self.same_folder_in_maintenance_log():
            state = Tk.NORMAL
        else:
            state = Tk.DISABLED
        self.plot_button.config( state=state )

    def same_folder_in_maintenance_log( self ):
        if self.maintenance_log is None:
            return False
        if not os.path.exists( self.maintenance_log ):
            return False

        fh = open( self.maintenance_log )
        line = fh.readline()
        fh.close()

        folder = line.replace( 'Doing ', '' )
        return folder[0:-1] == self.in_folder.get()

    def make_maintenance_log_path( self ):
        an_folder       = self.an_folder.get()
        analysis_name   = self.analysis_name.get()
        if is_empty_val(an_folder) or is_empty_val(analysis_name):
            return None

        return an_folder + '/' + analysis_name + '/.temp/maintenance.log'

    def save_settings( self, clear_temp=False ):
        restore_default_setting( 'file_extension' )
        if clear_temp:
            clear_temporary_settings()
        try:
            save_settings()
            molass_legacy.AutorgKek.Settings.save_settings()    # for saving AutorgKek settings
        except:
            pass

    def set_dev_setting( self, item, value ):
        set_dev_setting( item, value )

    def set_setting( self, item, value ):
        set_setting( item, value )

    def _exit(self):
        from molass_legacy.Optimizer.NpSharedMemory import close_shared_memory
        close_shared_memory()

        if self.cleaner is not None:
            if self.tmp_logger is None:
                # this can be avoided; temp fix
                pass
            else:
                self.tmp_logger.info("os._exit(0)")
            os._exit(0)

    def quit( self, immediately=False ):
        # print( 'quit' )

        if immediately:
            self.parent.quit()
            self.destroy()
            self._exit()
            return

        reply = MessageBox.askokcancel(
                "Quit Confirmation",
                "Do you really want to quit?",
                parent=self,
                )
        if not reply:
            return

        from molass_legacy.DENSS.DenssManagerDialog import terminate_manager
        ret = terminate_manager(self)
        if not ret:
            return

        from molass_legacy._MOLASS.Processes import terminate_all_processes
        terminate_all_processes()

        # print( self.geometry() )
        # set_setting( 'main_geometry',   self.geometry()          )

        """
        if self.settings_have_been_changed:
            reply = MessageBox.askyesno(
                    "Setting Info Save Question",
                    "Save the setting information?",
                    parent=self,
                    )
            if reply:
                save_settings()
        """
        save_settings(clear_no_save_items=True)
        self.update_setting_info( clear_temp=True )
        self.parent.quit()
        self.destroy()
        self._exit()

    def is_placed_foremost( self ):
        # TODO
        # num_stacked = len( self.tk.eval( 'wm stackorder '+str(self.top_widget) ).split(' ') )
        # return num_stacked == 1
        return True

    def _on_control_c( self, event ):
        w = self.focus_get()
        # print( 'table=', self.file_info_table )
        # print( 'w=', w )
        if str(w).find( str(self.file_info_table) ) == 0:
            self.file_info_table.table._on_control_c( event )
        else:
            # TODO:  w.selection
            pass

    def _on_click( self, event ):
        # the follwing order is essential to keep exluded marks (i.e., 'X')
        # unchanged in case of non-cell clicks.
        self.file_info_table.table._on_click( event )
        self.file_info_table._on_click( event )

    def _on_enter( self, event ):
        # TODO: only when changed
        # self.on_entry_in_folder()
        if not self.is_placed_foremost(): return

        print( 'ENTER in main window' )
        state = self.analysis_button.cget( 'state' )
        if state == Tk.NORMAL:
            self.analyze_all()
        else:
            self.in_folder_entry.on_entry()

    def _on_escape( self, event ):
        if not self.is_placed_foremost(): return

        self.quit()

    def do_simple_guinier( self, i ):
        from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
        from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter
        print( 'do_simple_guinier: i=', i )

        data, file, _, n = self.file_info_table.get_data_array( i )

        guinier = SimpleGuinier( data )
        adapter = AutorgKekAdapter( None, guinier=guinier )
        adapter_result = adapter.run()
        return adapter_result

    def show_threedim_dialog(self, debug=True):
        if debug:
            from importlib import reload
            import Tools.ThreeDimViewer
            reload(Tools.ThreeDimViewer)
        from molass_legacy.Tools.ThreeDimViewer import ThreeDimViewer
        self.threedim_dialog = ThreeDimViewer(self.parent, self.measured_data)
        self.threedim_dialog.show()

    def show_lpm_inspector(self):
        from molass_legacy.Baseline.LpmInspect import LpmInspector
        # from molass_legacy.Baseline.LpmInspect_for_paper import LpmInspector
        self.baseline_inspector = LpmInspector(self.parent, self.measured_data)
        self.baseline_inspector.show()

    def back_to_preproc(self):
        self.parent.after(0, self.back_to_preproc_impl)
        self.withdraw()

    def back_to_preproc_impl(self):
        if self.preproc_dialog is None:
            from Microfluidics.PreprocessDialog import PreprocessDialog
            self.preproc_dialog = PreprocessDialog(self.parent)
        self.preproc_dialog.show()
        self.update()
        self.deiconify()

    def disable_uv_data_tracer(self, *args):
        if not self.disable_uv_data_tracing:
            print('skipping disable_uv_data_tracer')
            return

        self.disable_uv_data_tracer_busy = True

        disable_uv_data = self.disable_uv_data.get()
        set_setting('disable_uv_data', disable_uv_data)
        if disable_uv_data:
            self.uv_folder.set('')
            self.uv_file.set('')
        self.uv_file_info_changed = True
        self.refresh(keeping_setting_items=['disable_uv_data'])

        self.disable_uv_data_tracer_busy = False

    def get_it_ready_for_qmm(self):
        try:
            analysis_name, analysis_folder = self.make_analysis_folder()
            if os.path.exists( analysis_folder ) and not is_almost_empty_dir( analysis_folder ):
                analysis_name, analysis_folder = self.new_analysis()
            else:
                self.analysis_name.set(analysis_name)   # TODO: find a better place to do this
                set_setting( 'analysis_folder', analysis_folder )
                set_setting( 'analysis_name', analysis_name )

        except:
            print(sys.exc_info())
            return False

        self.update_setting_info()
        self.analyzer.change_log_to(analysis_folder)
        return True

    def is_qmm_busy(self):
        return self.is_executing

    def get_qmm_controller(self):
        return self.qmm_controller

    def get_setting_for_test(self, item):
        return get_setting(item)

    def get_fullopt_dialog(self):
        return self.fullopt_dialog

    def get_v2_menu(self):
        return self.v2_menu
    
    def show_dev_trial_entry(self):
        from importlib import reload
        import Tools.DevTrial
        reload(Tools.DevTrial)
        from molass_legacy.Tools.DevTrial import dev_trial_entry
        dev_trial_entry(self)



"""

    ファイル名：   ThreadUtils.py

    Copyright (c) 2017-2023, SAXS Team, KEK-PF
"""

import ctypes

"""
    http://stackoverflow.com/questions/323972/is-there-any-way-to-kill-a-thread-in-python
"""
def terminate_thread(thread):
    """Terminates a python thread from another thread.

    :param thread: a threading.Thread instance
    """
    if not thread.is_alive():
        return

    exc = ctypes.py_object(SystemExit)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
        ctypes.c_long(thread.ident), exc)
    if res == 0:
        raise ValueError("nonexistent thread id")
    elif res > 1:
        # """if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"""
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread.ident, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")



"""
    IterativeLrfSolverDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt

def demo_impl(D, E, qv, x, y, f, t, data_title):
    from SvdDenoise import get_denoised_data
    from SimpleGuinier import SimpleGuinier
    from importlib import reload
    import BoundedLRF.IterativeLrfSolver
    reload(BoundedLRF.IterativeLrfSolver)
    from .IterativeLrfSolver import IterativeLrfSolver

    i = bisect_right(qv, 0.02)
    cy = D[i,:]
    j = np.argmax(cy)
    pv1 = D[:,j]
    pv1 = pv1/np.max(pv1)

    range_ = slice(f, t+1)

    solver = IterativeLrfSolver(qv, D[:,range_], E[:,range_])
    P, C, Rg, R, L, bq_bounds, coerced_bq, changes = solver.solve(maxiter=100000, L=0.4)

    with plt.Dp():
        cx = np.arange(len(changes))*100
        fig, ax = plt.subplots()
        ax.plot(cx, changes)
        fig.tight_layout()
        plt.show()

    aq, bq = P.T
    c1, c2 = C[:,j-f]
    aq_ = aq*c1
    bq_ = bq*c2
    coerced_bq *= c2

    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(18,10))
        ax1, ax2, ax3 = axes[0,:]
        fig.suptitle("Bounnded LRF Solver Demo for %s" % data_title, fontsize=20)
        ax1.set_title("Elution and Range", fontsize=16)
        ax2.set_title("LRF Result (Linear)", fontsize=16)
        ax3.set_title("LRF Result (Log)", fontsize=16)
        ax1.plot(x, y)
        ymin, ymax = ax1.get_ylim()
        p = Rectangle(
                (f, ymin),  # (x,y)
                t - f,   # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax1.add_patch(p)
        ax3.set_yscale('log')
        for ax in [ax2, ax3]:
            for k, pv in enumerate([pv1] + list(P.T)):
                if k < 2:
                    color = None
                    alpha = 1
                else:
                    color = "pink"
                    alpha = 0.5
                ax.plot(qv, pv, color=color, alpha=alpha)

        ax2_ = axes[1,1]
        ax2_.set_title("Bound-modified B(q)", fontsize=16)

        ax2_.plot(qv, aq_)
        ax2_.plot(qv, coerced_bq, color="pink", alpha=0.5)

        for ax in [ax2, ax2_]:
            for k, bv in enumerate(bq_bounds):
                label = "Bound L=%g" % L if k == 0 else None
                ax.plot(qv, bv, color="red", label=label, alpha=0.5)

            ax.set_ylim(-5, 5)
            ax.legend(fontsize=16)

        for ax in axes[1,0],axes[1,2]:
            ax.grid(False)
            ax.set_axis_off()

        fig.tight_layout()
        plt.show()

def demo(dialog, pdata, popts):
    from DataUtils import get_in_folder

    sd = dialog.sd
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    x = xr_curve.x
    y = xr_curve.y
    print("pdata.cnv_ranges=", pdata.cnv_ranges)
    paired_range = pdata.cnv_ranges[1]
    ranges = paired_range.get_fromto_list()
    print(ranges)
    f, t = ranges[1]
    data_title = get_in_folder()

    demo_impl(D, E, qv, x, y, f, t, data_title)

def demo2(caller):
    from importlib import reload

    import Tools.EmbedCushionUtils
    reload(Tools.EmbedCushionUtils)
    from molass_legacy.Tools.EmbedCushionUtils import get_caller_attr

    gi_in_folder = r"D:\PyTools\Data\20180526\GI"
    cushion_dict = get_caller_attr(caller, "cushion_dict", {})
    gi_sd = cushion_dict.get("gi_sd")
    if gi_sd is None:
        from molass_legacy.Tools.SdUtils import get_sd
        cushion_dict["gi_sd"] = gi_sd = get_sd(gi_in_folder)

    from DataUtils import get_in_folder

    sd = gi_sd
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    x = xr_curve.x
    y = xr_curve.y

    data_title = get_in_folder(gi_in_folder)
    f = 130
    t = 200
    demo_impl(D, E, qv, x, y, f, t, data_title)

def demo3(dialog, pdata, popts):
    from DataUtils import get_in_folder
    from importlib import reload
    import Trials.BoundedLRF.IterativeLrfSolver
    reload(Trials.BoundedLRF.IterativeLrfSolver)
    from .IterativeLrfSolver import IterativeLrfSolver

    sd = dialog.sd
    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    x = xr_curve.x
    y = xr_curve.y
    print("pdata.cnv_ranges=", pdata.cnv_ranges)
    paired_range = pdata.cnv_ranges[1]
    ranges = paired_range.get_fromto_list()
    print(ranges)
    f, t = ranges[1]
    data_title = get_in_folder()

    i = bisect_right(qv, 0.02)
    cy = D[i,:]
    j = np.argmax(cy)
    pv1 = D[:,j]
    # pv1 = pv1/np.max(pv1)

    range_ = slice(f, t+1)

    # L = 0.4
    L = 1
    bq_changes = []

    solver = IterativeLrfSolver(qv, D[:,range_], E[:,range_])
    P, C, Rg, R, L, bq_bounds, coerced_bq, w, changes = solver.solve(maxiter=1, L=L)
    c1, c2 = C[:,j-f]
    aq, bq = P.T
    aq1 = aq*c1
    bq1 = bq
    bq_changes.append((C, bq1, coerced_bq, bq_bounds, w))

    P, C, Rg, R, L, bq_bounds, coerced_bq, w, changes = solver.solve(maxiter=100, L=L)
    c1, c2 = C[:,j-f]
    aq, bq = P.T
    aq_ = aq*c1
    bq_ = bq
    bq_changes.append((C, bq_, coerced_bq, bq_bounds, w))

    P, C, Rg, R, L, bq_bounds, coerced_bq, w, changes = solver.solve(maxiter=1000, L=L)
    c1, c2 = C[:,j-f]
    aq, bq = P.T
    aq_ = aq*c1
    bq_ = bq
    bq_changes.append((C, bq_, coerced_bq, bq_bounds, w))

    P, C, Rg, R, L, bq_bounds, coerced_bq, w, changes = solver.solve(maxiter=100000, L=L)
    c1, c2 = C[:,j-f]
    aq, bq = P.T
    aq_ = aq*c1
    bq_ = bq
    bq_changes.append((C, bq_, coerced_bq, bq_bounds, w))

    with plt.Dp():
        fig, axes = plt.subplots(nrows=3, ncols=4, figsize=(20,10))
        ax1, ax2, ax3, ax4 = axes[0,:]
        fig.suptitle("Bounnded LRF Solver Demo for %s with L=%g" % (data_title, L), fontsize=20)
        ax1.set_title("Elution Data and Range", fontsize=16)
        ax2.set_title("LRF Result (Linear)", fontsize=16)
        ax3.set_title("LRF Result (Log)", fontsize=16)
        ax4.set_title("LRF Result (Guinier Plot)", fontsize=16)
        ax1.plot(x, y)
        ymin, ymax = ax1.get_ylim()
        p = Rectangle(
                (f, ymin),  # (x,y)
                t - f,   # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax1.add_patch(p)

        pv1 = pv1/pv1[i]
        aq1 = aq1/aq1[i]
        aqf = aq_/aq_[i]

        ax3.set_yscale("log")
        for ax in ax2, ax3:
            ax.plot(qv, pv1, label="data")
            ax.plot(qv, aq1, label="LRF first A(q)", alpha=0.5)
            ax.plot(qv, aqf, label="LRF final A(q)", alpha=0.5)
            ax.legend()
            axt = ax.twinx()
            axt.grid(False)
            axt.plot(qv, w, ":", color="cyan")

        glim_i = bisect_right(qv, 1.5/Rg)
        gslice = slice(0,glim_i)
        qv2 = qv[gslice]**2

        ax4.plot(qv2, np.log(pv1[gslice]), label="data")
        ax4.plot(qv2, np.log(aq1[gslice]), label="LRF first A(q)", alpha=0.5)
        ax4.plot(qv2, np.log(aqf[gslice]), label="LRF final A(q)", alpha=0.5)
        ax4.legend()

        axc_ylims = []
        ax_ylims = []
        x_ = x[range_]
        for axc, ax, (C, bq_, coerced_bq, bq_bounds, w) in zip(axes[1,:], axes[2,:], bq_changes):
            axc.plot(x, y)
            for k, cv in enumerate(C):
                color = "pink" if k == 1 else None
                axc.plot(x_, cv, color=color)
            axc_ylims.append(axc.get_ylim())
            ax.plot(qv, bq_, color="pink", label="B(q)")
            ax.plot(qv, coerced_bq, color="yellow", alpha=0.5, label="coerced B(q)")
            ax.plot(qv, bq_bounds[0], ":", color="red", alpha=0.5, label="lower bound")
            ax.plot(qv, bq_bounds[1], ":", color="red", alpha=0.5, label="upper bound")
            ax.set_ylim(-2, 2)
            ax.legend()
            ax_ylims.append(ax.get_ylim())

        for axes_, ylims in [(axes[1,:], axc_ylims), (axes[2,:], ax_ylims)]:
            ylims = np.array(ylims)
            ymin = np.min(ylims[:,0])
            ymax = np.max(ylims[:,1])
            for ax in axes_:
                ax.set_ylim(ymin, ymax)

        fig.tight_layout()
        plt.show()



"""

    AnalysisGuide.py

    Copyright (c) 2019-2023, SAXS Team, KEK-PF

"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Mapping.SingleComponent import ACCEPTABLE_MIN_SCI, PEAK_EVAL_RANGE_RATIO

class AnalysisGuide:
    def __init__(self, mapper, judge_holder):
        self.mapper = mapper
        self.judge_holder = judge_holder

    def already_decomposed(self):
        use_elution_models = get_setting('use_elution_models')
        decomp_editor_info = get_setting('decomp_editor_info')
        return use_elution_models and decomp_editor_info is not None

    def has_bad_sci(self):
        min_sci = self.mapper.get_min_sci()
        print('min_sci=', min_sci)
        return min_sci < ACCEPTABLE_MIN_SCI, 'a peak with SCI %d less than %d' % (min_sci, ACCEPTABLE_MIN_SCI), None

    def has_overlapping_peaks(self, debug=False):
        """
        TODO:
            better sigma points by considering asymmetry
        """
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            x = self.mapper.x_curve.x
            y = self.mapper.x_curve.y
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(x, y)
            ax.set_ylim(ax.get_ylim())

        bool_ret = False

        peaks = self.mapper.x_curve.get_emg_peaks()
        prev_point = None
        max_length = 0
        for k, peak in enumerate(peaks):
            try:
                points = peak.get_model_x_from_ratio(PEAK_EVAL_RANGE_RATIO)
            except:
                import logging
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                logger = logging.getLogger(__name__)
                etb = ExceptionTracebacker()
                logger.warning("failed to get foot points:" + str(etb))
                continue
            if debug:
                ax.plot(x, peak.get_model_y(x))
                ymin, ymax = ax.get_ylim()
                for p in points:
                    ax.plot([p, p], [ymin, ymax], ':', color='gray')
            if prev_point is not None:
                length = prev_point - points[0]
                if length > max_length:
                    max_length = length
                    bool_ret = True
            prev_point = points[1]

        if debug:
            plt.show()

        return bool_ret, 'an overlapping interval of length %.3g' % max_length, None

    def has_uncomputable_rdrs(self):
        assert self.judge_holder is not None
        judge = self.judge_holder.has_uncomputable_rdrs()
        if judge:
            ret_message = "an uncomputable RDR"
            extra_info = "uncomputable"
        else:
            ret_message = "no uncomputable RDRs"
            extra_info = None
        return judge, ret_message, extra_info

    def need_decomp(self):
        reasons = []
        extra_infos = []
        ret_judge = False
        for judge, reason, extra_info in [self.has_bad_sci(), self.has_overlapping_peaks(), self.has_uncomputable_rdrs()]:
            ret_judge = ret_judge or judge
            if judge:
                reasons.append(reason)
            if extra_info is not None:
                extra_infos.append(extra_info)
        return ret_judge, ' and '.join(reasons), extra_infos

    def get_decomp_guide_message(self, reason):
        return  ( "Use of Decomposition Editor is reccommended\n"
                + " due to %s." % reason
                )

class AnalysisGuideInfo:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def decomp_proc_needed(self):
        # for the Tester to be able to call this
        return self.need_decomp and not self.decomposed

    def has_an_uncomputable_rdr(self):
        return len(self.extra_infos) > 0 and self.extra_infos[0] == "uncomputable"

def get_analysis_guide_info(mapper, judge_holder, discomfort):
    std_diff = mapper.std_diff

    if std_diff < 0.3:
        # TODO: refactor to unify 0.2
        if std_diff < 0.2:
            mapping_quality = "The mapping seems fairly good"
        else:
            mapping_quality = "The mapping seems acceptable"
    else:
        mapping_quality = "Some adjustment is recommended because the mapping is not so good."

    if discomfort < 0.1:
        three_d_message = ""
        three_d_guide = False
    else:
        three_d_message = "View the Xray scattering in 3D and consider baseline correction."
        three_d_guide = True

    guide = AnalysisGuide(mapper, judge_holder)

    decomposed = guide.already_decomposed()
    need_decomp, reason, extra_infos = guide.need_decomp()

    if need_decomp:

        message = guide.get_decomp_guide_message(reason)
        fg = 'orange'
        edit_mode = True

    else:

        if std_diff < 0.3:
            if discomfort < 0.1:
                message = mapping_quality + ". We can proceed by pressing [▶ Execute]."
                fg = 'green'
                edit_mode = False
            else:
                message = ( "Although " + mapping_quality.lower() + ", Xray scattering base discomfort is observed.\n"
                            + three_d_message
                            )
                fg = 'orange'
                edit_mode = True
        else:
            if discomfort < 0.1:
                message = mapping_quality
            else:
                message = ( mapping_quality + "\n"
                            + "Also, " + three_d_message.lower()
                            )
            fg = 'orange'
            edit_mode = True

    return AnalysisGuideInfo(
            message=message,
            fg=fg,
            edit_mode=edit_mode,
            three_d_guide=three_d_guide,
            decomposed=decomposed,
            need_decomp=need_decomp,
            extra_infos=extra_infos)

def get_data_correction_state_text():
    from molass_legacy.Mapping.MappingParams import get_mapper_opt_params

    opt_params = get_mapper_opt_params()

    uv_text = "UV data correction: " + opt_params.get_uv_correction_str()
    xr_text = "Xray data correction: " + opt_params.get_xray_correction_str()

    return uv_text + "; " + xr_text



# coding: utf-8
"""
    EmgBySampling.py

    Copyright (c) 2020,2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as mplt
from matplotlib.gridspec import GridSpec
import matplotlib.animation as animation
from matplotlib.patches import Polygon
from .AnimatedHistogram import AnimatedHistogram
import molass_legacy.KekLib.DebugPlot as plt

FONTSIZE = 16

class EmgBySampling:
    def __init__(self, num_bins=100, num_frames=200):
        self.num_bins = num_bins
        self.num_frames = num_frames

    def play(self, save_only=False):
        fig = plt.figure(figsize=(18, 9))
        gs = GridSpec(3,17)
        axes = []
        for i in range(3):
            ax1 = fig.add_subplot(gs[i,0:10])
            ax2 = fig.add_subplot(gs[i,10])
            ax3 = fig.add_subplot(gs[i,11:17])
            axes.append([ax1, ax2, ax3])
        self.axes = axes = np.array(axes)

        N = 10000
        self.cycle = N//self.num_frames
        X = np.random.normal(0, 1, N)
        Y = np.random.exponential(1, N)
        Z = X + Y
        rect1, bar1, _, ah1 = self.plot_data(*axes[0,:], X, "X: Gaussian")
        rect2, bar2, _, ah2 = self.plot_data(*axes[1,:], Y, "Y: Exponential", bar_color="pink")
        rect3, bar3, bar3_, ah3 = self.plot_data(*axes[2,:], Z, "X+Y: EMG", y=Y)
        rects = [rect1, rect2, rect3]
        bars = [bar1, bar2, bar3, bar3_]
        a_hists = [ah1, ah2, ah3]
        hist_patches = [ah.reset()[0] for ah in a_hists]

        plt.dp.mpl_canvas.draw()

        u_ymin, u_ymax = 999, -999
        for ax1 in axes[:,0]:
            ymin, ymax = ax1.get_ylim()
            u_ymin = min(u_ymin, ymin)
            u_ymax = max(u_ymax, ymax)
        for ax_ in axes[:,0:2]:
            for ax in ax_:
                ax.set_ylim(u_ymin, u_ymax)

        u_xmin, u_xmax = 999, -999
        for ax3 in axes[:,2]:
            xmin, xmax = ax3.get_xlim()
            u_xmin = min(u_xmin, xmin)
            u_xmax = max(u_xmax, xmax)
        for ax3 in axes[:,2]:
            ax3.set_xlim(u_xmin, u_xmax)

        fig.tight_layout()

        def reset():
            return rects + bars + hist_patches

        def animate(i):
            i_ = i * self.cycle
            for rect, x in zip(rects, [X, Y, Z]):
                new_vertices = [(0,u_ymin), (i_, u_ymin), (i_, u_ymax), (0, u_ymax)]
                rect.set_xy(new_vertices)

            for bar, x in zip(bars, [X, Y, Y]):
                bar.set_data([0.5, 0.5], [0, x[i_]])

            bar3_.set_data([0.5, 0.5], [Y[i_], Z[i_]])

            for ah in a_hists:
                ah.animate(i)

            return rects + bars + hist_patches

        self.anim = animation.FuncAnimation(fig, animate, frames=self.num_frames, blit=True, init_func=reset, interval=100)

        if save_only:
            pass
        else:
            plt.show()

    def plot_data(self, ax1, ax2, ax3, x, dtype_title, y=None, bar_color=None):
        N = len(x)
        title_with_num = dtype_title.replace(':', ': ' + str(N) )
        ax1.set_title(title_with_num + " Random Values", fontsize=FONTSIZE)
        ax2.set_title("Bar", fontsize=FONTSIZE)
        ax3.set_title(dtype_title + " Histogram", fontsize=FONTSIZE)
        ax1.plot(x, 'o', markersize=1)
        m = np.mean(x)
        s = np.std(x)
        px = [0, N-1]
        line_stype = '-'
        ax1.plot(px, [m, m], line_stype, color='green', linewidth=3, label='mean')
        ax1.plot(px, [m-s, m-s], line_stype, color='orange', linewidth=3, label='mean-stdev')
        ax1.plot(px, [m+s, m+s], line_stype, color='orange', linewidth=3, label='mean+stdev')
        ax1.plot(px, [0, 0], ':', color='red', linewidth=3, label='zero')
        ax1.legend(loc='upper right')

        vertices = [(0,0), (1,0), (1,1), (0,1)]
        rect = Polygon(vertices, color='yellow', alpha=0.2)
        ax1.add_patch(rect)
        bar, = ax2.plot([0.5, 0.5], [0, 0], solid_capstyle='butt', linewidth=10, color=bar_color)
        if y is None:
            bar_ = None
        else:
            bar_, = ax2.plot([0.5, 0.5], [0, 0], solid_capstyle='butt', linewidth=10, color='pink')
        ax2.plot([0, 1], [0, 0], ':', color='red', linewidth=3)
        ax2.set_xlim(0, 1)
        ax2.get_xaxis().set_visible(False)

        patch_options = [{'facecolor':None}, {'facecolor':'yellow', 'alpha':0.3}]
        ah = AnimatedHistogram(x, self.num_bins, self.num_frames, patch_options=patch_options)
        ah.prepare(ax3)
        ymin, ymax = ax3.get_ylim()
        ax3.set_ylim(ymin, ymax)
        ax3.plot([m,m], [ymin, ymax], line_stype, color='green', linewidth=3, label='mean')
        ax3.plot([m-s,m-s], [ymin, ymax], line_stype, color='orange', linewidth=3, label='mean-stdev')
        ax3.plot([m+s,m+s], [ymin, ymax], line_stype, color='orange', linewidth=3, label='mean+stdev')
        ax3.plot([0,0], [ymin, ymax], ':', color='red', linewidth=3, label='zero')
        ax3.legend(loc='upper right')

        return rect, bar, bar_, ah

    def save(self, file, **kwargs):
        self.anim.save(file, **kwargs)

def moments_proof():
    N = 1000000

    m1_list = []
    m2_list = []
    m3_list = []
    x = np.arange(0.1, 3, 0.1)
    for tau in x:
        print('tau=', tau)
        X = np.random.normal(0, 1, N)
        Y = np.random.exponential(tau, N)
        Z = X + Y

        M1 = np.mean(Z)
        print('M1=', M1, tau)
        m1_list.append((M1, tau))
        M2 = np.sum((Z - M1)**2)/N
        print('M2=', M2, 1+tau**2)
        m2_list.append((M2, 1+tau**2))

        M3 = np.sum((Z - M1)**3)/N
        print('M3=', M3, 2*tau**3)
        m3_list.append((M3, 2*tau**3))

    m1_array = np.array(m1_list)
    m2_array = np.array(m2_list)
    m3_array = np.array(m3_list)

    fig = plt.figure()
    ax = fig.gca()
    ax.set_title("Proof of EMG Moment Formulas with %d Sampling" % N, fontsize=16)
    ax.set_xlabel(r'$\tau$')
    ax.set_ylabel('Moments')
    ax.plot(x, m1_array[:,0], color='red', label=r'$M_1$ (raw)')
    ax.plot(x, m1_array[:,1], ':', color='red', label=r'$\mu + \tau$')
    ax.plot(x, m2_array[:,0], color='green', label=r'$M_2$ (central)')
    ax.plot(x, m2_array[:,1], ':', color='green', label=r'$\mu^2 + \tau^2$')
    ax.plot(x, m3_array[:,0], color='blue', label=r'$M_3$ (central)')
    ax.plot(x, m3_array[:,1], ':', color='blue', label=r'$2\tau^3$')
    ax.legend()

    fig.tight_layout()
    plt.show()



"""
    LogGuinierKratkyPlotter.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.patches as mpl_patches
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier

class LogGuinierKratkyPlotter(Dialog):
    def __init__(self, parent, data, path):
        self.data = data
        self.path = path
        Dialog.__init__( self, parent, "Log Guinier Kratky Plotter", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        from importlib import reload
        import GuinierAnalyzer.GuinierKratkyPlots
        reload(GuinierAnalyzer.GuinierKratkyPlots)
        from molass_legacy.GuinierAnalyzer.GuinierKratkyPlots import guinier_kratky_plots_impl

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        sg = SimpleGuinier(self.data)
        qv, y, e = self.data.T

        fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(18,5))
        fig.suptitle("LGK Plot for %s" % self.path, fontsize=20)

        ax1.set_title("Log Plot", fontsize=16)
        ax1.set_yscale('log')
        ax1.plot(qv, y)

        interval = sg.guinier_start, sg.guinier_stop
        color = None
        guinier_kratky_plots_impl(ax2, ax3, qv, y, sg.Rg, sg.Iz, color, interval=interval, markersize=1)

        fig.tight_layout()
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
